is_editor = true
local es = {}


require("window", true)
require "game"
require "player"
require "level"
require "level_visuals"

local save_index

local
function load_json(filename)
    local f = io.open(filename)
    if f then
        local json = f:read("*a")
        f:close()
        return am.parse_json(json)
    else
        return nil
    end
end

local
function save_json(filename, val)
    local f = io.open(filename, "w")
    f:write(am.to_json(val))
    f:close()
end

local
function load_save_index()
    save_index = load_json("editor_data/save_index.json")
end

load_save_index()

local
function save_save_index() 
    save_json("editor_data/save_index.json", save_index)
end

local editor_sprites = require "editor_sprites"

local timeline_frames = 200
local frame_secs = 0.1
local frame_width = 18
local timeline_height = 45
local timeline_bg_color = vec4(0.75, 0.75, 0.75, 1)
local timeline_notch_color = vec4(0.5, 0.5, 0.5, 1)
local timeline_divider_color = vec4(0, 0, 0, 1)
local timeline_divider_width = 1
local timeline_top = win.bottom + timeline_height
local keyframe_color = vec4(1, 0.75, 0.0, 1)
local snap_xy = 4
local snap_angle = math.rad(5)
local tile_sz = 32
local transparent = vec4(0)
local filled_tile = vec4(0.2, 0.8, 0, 1)
local grid_line_thickness = 2
local grid_line_color = vec4(0.0, 0.0, 0.5, 1)
local origin_cursor_color = vec4(1.0, 1.0, 0.0, 1)
local edit_area_top = win.top - 90
local edit_area_bottom = win.bottom + 90
local default_layer_w = 64
local default_layer_h = 64
local level_name_color = vec4(1)
local modified_name_color = vec4(1, 0.3, 0.3, 1)

local disabled_shader = am.program(
[[
    precision highp float;
    attribute vec2 vert;
    attribute vec2 uv;
    uniform mat4 MV;
    uniform mat4 P;
    varying vec2 v_uv;
    void main() {
        v_uv = uv;
        gl_Position = P * MV * vec4(vert, 0.0, 1.0);
    }
]],
[[
    precision mediump float;
    uniform sampler2D tex;
    uniform vec4 color;
    varying vec2 v_uv;
    void main() {
        vec4 sample = texture2D(tex, v_uv);
        float lum = (sample.r + sample.g + sample.b) * sample.a / 3.0;
        gl_FragColor = vec4(vec3(lum), sample.a) * color;
    }
]]
)

local toggled_shader = am.program(
[[
    precision highp float;
    attribute vec2 vert;
    attribute vec2 uv;
    uniform mat4 MV;
    uniform mat4 P;
    varying vec2 v_uv;
    void main() {
        v_uv = uv;
        gl_Position = P * MV * vec4(vert, 0.0, 1.0);
    }
]],
[[
    precision mediump float;
    uniform sampler2D tex;
    uniform vec4 color;
    varying vec2 v_uv;
    void main() {
        vec4 sample = texture2D(tex, v_uv);
        gl_FragColor = vec4(1.0 - sample.rgb, sample.a) * color;
    }
]]
)

local layer_shader = am.program(
[[
    precision highp float;
    attribute vec2 vert;
    uniform mat4 MV;
    uniform mat4 P;
    void main() {
        gl_Position = P * MV * vec4(vert, 0.0, 1.0);
    }
]],
[[
    precision mediump float;
    uniform vec4 tint;
    uniform vec4 color;
    void main() {
        gl_FragColor = color * tint;
    }
]]
)

local
function snap(x, dx)
    return math.floor(x / dx + 0.5) * dx
end

local
function snap_scale(x)
    return snap(x, 1/64)
end

local button_toggle_groups = {}

local
function create_button(sprite, x, y, on_press, toggle_group)
    local h, w = sprite.width, sprite.height
    local x_start = x
    local x_end = x_start + w
    local y_top = y + h
    local y_bottom = y
    local group = am.group{
        am.translate(x, y) ^ am.sprite(sprite, "left", "bottom")
    }
    local disabled = false
    local toggled = false
    group:action(function()
        if not disabled and input_pressed() then
            local pos = input_pos()
            if pos.x > x_start and pos.x < x_end and pos.y > y_bottom and pos.y < y_top then
                if toggle_group then
                    if button_toggle_groups[toggle_group] then
                        button_toggle_groups[toggle_group].toggled = false
                    end
                    button_toggle_groups[toggle_group] = group
                    group.toggled = true
                end
                on_press()
            end
        end
    end)
    function group:get_disabled()
        return disabled
    end
    local
    function set_shader()
        if disabled then
            group"use_program".program = disabled_shader
        elseif toggled then
            group"use_program".program = toggled_shader
        else
            group"use_program".program = am.shaders.premult_texturecolor2d
        end
    end

    function group:set_disabled(val)
        disabled = not not val
        set_shader()
    end
    if toggle_group then
        function group:get_toggled()
            return toggled
        end
        function group:set_toggled(val)
            toggled = not not val
            set_shader()
        end
    end
    return group
end

local
function get_timeline_notch_color(f)
    return f % 5 == 0 and timeline_notch_color or timeline_bg_color
end

local
function create_timeline_quads()
    local quads = am.quads(1, {"vert", "vec2", "color", "vec4"})

    -- frame cells
    local frame_quads = {}
    local x = 0
    local y1 = win.bottom
    local y2 = y1 + timeline_height
    for f = 1, timeline_frames do
        local color = get_timeline_notch_color(f)
        frame_quads[f] = quads:add_quad{
            vert = {x, y2, x, y1, x + frame_width, y1, x + frame_width, y2},
            color = color,
        }
        x = x + frame_width
    end

    -- dividing lines
    x = 0
    for f = 1, timeline_frames do
        quads:add_quad{
            vert = {x, y2, x, y1, x + timeline_divider_width, y1, x + timeline_divider_width, y2},
            color = timeline_divider_color,
        }
        x = x + frame_width
    end
    
    return quads, frame_quads
end

local
function get_frame(keyframes, frame)
    local prev_kf
    for f = frame, 1, -1 do
        if keyframes[f] then
            prev_kf = f
            break
        end
    end
    if prev_kf == frame then
        return keyframes[frame]
    end
    -- find next keyframe
    local next_kf
    for f = frame + 1, #keyframes do
        if keyframes[f] then
            next_kf = f
            break
        end
    end
    if not next_kf and prev_kf then
        return keyframes[prev_kf]
    elseif next_kf and not prev_kf then
        return keyframes[next_kf]
    elseif not next_kf and not prev_kf then
        return {
            x = 0,
            y = 0,
            angle = 0,
            scale_x = 1,
            scale_y = 1,
            origin_x = 0,
            origin_y = 0,
            ease = 1,
        }
    end
    local values1 = keyframes[prev_kf]
    local values2 = keyframes[next_kf]
    local tween_frame = {}
    local r = layer_easings[values1.ease]((frame - prev_kf) / (next_kf - prev_kf))
    for k, v in pairs(values1) do
        if k ~= "ease" then
            tween_frame[k] = (1 - r) * values1[k] + r * values2[k]
        end
    end
    tween_frame.ease = values1.ease
    return tween_frame
end

local
function get_prev_keyframe(keyframes, frame)
    for f = frame, 1, -1 do
        if keyframes[f] then
            return keyframes[f]
        end
    end
    return nil
end

local
function get_or_create_keyframe(keyframes, frame)
    local kf = get_frame(keyframes, frame)
    if kf ~= keyframes[frame] then
        keyframes[frame] = table.shallow_copy(kf)
    end
    return keyframes[frame]
end

local
function update_player_pos()
    local recording = es.player_recording
    if not recording then
        es.player_pos_node.hidden = true
        return
    end
    es.player_pos_node.hidden = false
    local t = (es.current_frame - 1) * settings.time_per_frame
    local i = #recording
    while i >= 1 and recording[i].time > t do
        i = i - 1
    end
    if i >= 1 then
        es.player_pos_node.x = recording[i].x
        es.player_pos_node.y = recording[i].y
    else
        es.player_pos_node.hidden = true
    end
end

local 
function compute_layer_transform(layer, n)
    if not layer.transform_dirty then
        return
    end
    local parent_mat = mat4(1)
    if (n or 0) > 10 then
        return
    end
    if layer.parent then
        compute_layer_transform(es.layers[layer.parent], (n or 0) + 1)
        parent_mat = es.layers[layer.parent].transform
    end
    local frame = get_frame(layer.keyframes, es.current_frame)
    layer.transform = parent_mat * 
        math.translate4(frame.x + frame.origin_x, frame.y + frame.origin_y) *
        math.rotate4(frame.angle) *
        math.scale4(frame.scale_x, frame.scale_y) *
        math.translate4(-frame.origin_x, -frame.origin_y)
    layer.transform_dirty = false
    layer.transform_node.mat = layer.transform
end

local
function update_layer_info()
    local layer = es.layers[es.active_layer]
    local frame = get_frame(layer.keyframes, es.current_frame)
    if es.active_layer == 1 then
        es.layer_info_node"text".text = string.format(
    [[
    Layer:    %d (GRID)
    Parent:   %s
    pos:      %0.2f %0.2f
    origin:   %0.2f %0.2f
    angle:    %d (%d)
    scale:    %0.3f %0.3f]], 
        es.active_layer, 
        layer.parent or "",
        frame.x, frame.y,
        frame.origin_x, frame.origin_y,
        math.deg(frame.angle), math.deg(frame.angle) % 360,
        frame.scale_x, frame.scale_y)
    else
        es.layer_info_node"text".text = string.format(
    [[
    Layer:    %d
    Parent:   %s
    pos:      %0.2f %0.2f
    origin:   %0.2f %0.2f
    angle:    %d (%d)
    scale:    %0.3f %0.3f
    Stencil:  %s
    sHader:   %s]], 
        es.active_layer, 
        layer.parent or "",
        frame.x, frame.y,
        frame.origin_x, frame.origin_y,
        math.deg(frame.angle), math.deg(frame.angle) % 360,
        frame.scale_x, frame.scale_y,
        layer.stencil or "always",
        layer_shaders[layer.shader].name)
    end
end

local
function update_layers()
    for l, layer in ipairs(es.layers) do
        layer.transform_dirty = true
    end
    for l, layer in ipairs(es.layers) do
        compute_layer_transform(layer)
        local frame = get_frame(layer.keyframes, es.current_frame)
        layer.origin_cursor.position2d = vec2(frame.origin_x, frame.origin_y)
    end
    update_player_pos()
    update_layer_info()
end

local
function update_ease_buttons()
    local keyframes = es.layers[es.active_layer].keyframes
    local frame = get_frame(keyframes, es.current_frame)
    if es.ease ~= frame.ease then
        button_toggle_groups.ease.toggled = false
        es.ease = frame.ease
        button_toggle_groups.ease = es.ease_buttons[es.ease]
        button_toggle_groups.ease.toggled = true
    end
end

local
function update_timeline()
    local x = -(es.current_frame - 1) * frame_width
    es.timeline_translate.position2d = vec2(x, 0)
    es.current_frame_text.text = es.current_frame
    update_layers()
    update_ease_buttons()
end

local
function update_timeline_for_active_layer()
    local keyframes = es.layers[es.active_layer].keyframes
    for f = 1, timeline_frames do
        local quad = es.frame_quads[f]
        if keyframes[f] then
            es.timeline_quads:quad_color(quad, keyframe_color)
        else
            es.timeline_quads:quad_color(quad, get_timeline_notch_color(f))
        end
    end
end

local
function active_layer_change()
    update_timeline_for_active_layer()
    update_timeline()
    for l, layer in ipairs(es.layers) do
        if l == es.active_layer then
            layer.node"bind".tint = vec4(1)
        else
            layer.node"bind".tint = vec4(1, 1, 1, 0.3)
        end
    end
    update_layer_info()
end

local
function update_level_name()
    es.level_name_text.text = es.level_name.." v"..es.level_version..(es.modified and "*" or "")
    es.level_name_text.color = es.modified and modified_name_color or level_name_color
end

local
function init_keyframes()
    local keyframes = {}
    for i = 1, timeline_frames do
        keyframes[i] = false
    end
    return keyframes
end

local
function find_active_layer_cell(pos)
    if es.active_layer == 1 then
        return nil
    end
    local layer = es.layers[es.active_layer]
    local frame = get_frame(layer.keyframes, es.current_frame)
    local num_rows = layer.height
    local num_columns = layer.width
    local layer_pos = (math.inverse(layer.transform) * vec4(pos, 0, 1)).xy
    local column = math.ceil(layer_pos.x / tile_sz + num_columns / 2)
    local row = math.ceil(layer_pos.y / tile_sz + num_rows / 2)
    if column < 1 or row < 1 or column > num_columns or row > num_rows then
        return nil
    else
        return row, column
    end
end

local
function refresh_layer_mesh(layer)
    local tris = compute_layer_tris(layer)
    if #tris == 0 then
        layer.mesh.vert = am.vec2_array{0, 0, 0, 0, 0, 0}
    else
        layer.mesh.vert = am.vec2_array(compute_layer_tris(layer))
    end
end

local
function create_origin_cursor()
    local origin_quads = am.quads(2, {"vert", "vec2"})
    local w = 10
    local h = 1
    origin_quads:add_quad{
        vert = {-w, h, -w, -h, w, -h, w, h},
    }
    origin_quads:add_quad{
        vert = {-h, w, -h, -w, h, -w, h, w},
    }
    return am.translate(0, 0) ^ am.bind{color = origin_cursor_color} ^ origin_quads
end

local
function gen_grid_crosshairs(group, rows, cols)
    local positions = gen_grid_positions(rows, cols)
    for i = 1, rows do
        for j = 1, cols do
            local x = positions[i][j].x
            local y = positions[i][j].y
            local line = am.line(vec2(x-5, y), vec2(x+5, y), 3)
            group:append(line)
            line = am.line(vec2(x, y-5), vec2(x, y+5), 3)
            group:append(line)
        end
    end
end

local
function create_grid_layer(rows, cols)
    local grid = am.group()
    gen_grid_crosshairs(grid, rows, cols)
    local origin_cursor = create_origin_cursor()
    local node = 
        am.transform(mat4(1))
        ^ am.bind{tint = vec4(1)}
        ^ am.use_program(layer_shader)
        ^ am.blend"alpha"
        ^ {
            grid,
            origin_cursor,
        }
    local layer = {
        node = node,
        transform_node = node,
        transform = mat4(1),
        keyframes = init_keyframes(),
        origin_cursor = origin_cursor,
        grid_group = grid,
    }
    return layer
end

local
function regen_grid()
    local layer = es.layers[1]
    local group = layer.grid_group
    group:remove_all()
    gen_grid_crosshairs(group, es.grid_rows, es.grid_cols)
end

local
function create_layer(width, height)
    local tiles = {}
    for i = 1, height do
        tiles[i] = {}
        for j = 1, width do
            tiles[i][j] = 0
        end
    end

    -- grid lines
    local grid_quads = am.quads(1, {"vert", "vec2"})
    local t = grid_line_thickness / 2
    local x1 = -width * tile_sz / 2
    local x2 = width * tile_sz / 2
    y = -height * tile_sz / 2
    for i = 0, height do
        grid_quads:add_quad{
            vert = {x1, y + t, x1, y - t, x2, y - t, x2, y + t},
        }
        y = y + tile_sz
    end
    local y1 = -height * tile_sz / 2
    local y2 = height * tile_sz / 2
    local x = -width * tile_sz / 2
    for j = 0, width do
        grid_quads:add_quad{
            vert = {x - t, y2, x - t, y1, x + t, y1, x + t, y2},
        }
        x = x + tile_sz
    end

    local origin_cursor = create_origin_cursor()
    local mesh = am.bind{
        vert = am.vec2_array{0, 0, 0, 0, 0, 0}
    } ^ am.draw"triangles"
    local node = am.transform(mat4(1))
        ^ am.bind{tint = vec4(1)}
        ^ am.use_program(layer_shader)
        ^ am.blend"alpha"
        ^ {
            am.bind{ color = vec4(0, 1, 0, 1) }
            ^ mesh
            ,
            am.bind{ color = grid_line_color }
            ^ grid_quads
            ,
            origin_cursor,
        }
    local layer = {
        node = node,
        mesh = mesh,
        origin_cursor = origin_cursor,
        transform_node = node,
        tiles = tiles,
        keyframes = init_keyframes(),
        width = width,
        height = height,
        shader = 1,
        transform_dirty = true,
    }
    refresh_layer_mesh(layer)
    return layer
end

local
function goto_start()
    local start
    for l, layer in ipairs(es.layers) do
        for f = 1, #layer.keyframes do
            if layer.keyframes[f] then
                if not start or f < start then
                    start = f
                end
                break
            end
        end
    end
    start = start or 1
    es.current_frame = start
    update_timeline()
end

local
function goto_end()
    local last
    for l, layer in ipairs(es.layers) do
        for f = #layer.keyframes, 1, -1 do
            if layer.keyframes[f] then
                if not last or f > last then
                    last = f
                end
                break
            end
        end
    end
    last = last or 1
    es.current_frame = last
    update_timeline()
end

local
function cycle_active_layer()
    es.active_layer = es.active_layer % #es.layers + 1
    active_layer_change()
end

local
function cycle_layer_parent()
    local layer = es.layers[es.active_layer]
    if not layer.parent then
        layer.parent = 1
    else
        layer.parent = layer.parent + 1
        if layer.parent > #es.layers then
            layer.parent = nil
        end
    end
    if layer.parent == es.active_layer then
        layer.parent = layer.parent + 1
        if layer.parent > #es.layers then
            layer.parent = nil
        end
    end
    es.modified = true
    update_layers()
end

local
function cycle_layer_shader()
    if es.active_layer == 1 then
        return
    end
    local layer = es.layers[es.active_layer]
    layer.shader = layer.shader % #layer_shaders + 1
    es.modified = true
    update_layers()
end

local
function cycle_layer_stencil()
    if es.active_layer == 1 then
        return
    end
    local layer = es.layers[es.active_layer]
    if not layer.stencil then
        layer.stencil = "zero"
    elseif layer.stencil == "zero" then
        layer.stencil = "one"
    elseif layer.stencil == "one" then
        layer.stencil = "never"
    elseif layer.stencil == "never" then
        layer.stencil = nil
    end
    es.modified = true
    update_layers()
end

local
function move_layer_up()
    if es.active_layer == #es.layers or es.active_layer == 1 then
        return
    end
    es.layers[es.active_layer], es.layers[es.active_layer + 1] = 
        es.layers[es.active_layer + 1], es.layers[es.active_layer]
    for _, layer in ipairs(es.layers) do
        if layer.parent == es.active_layer then
            layer.parent = es.active_layer + 1
        elseif layer.parent == es.active_layer + 1 then
            layer.parent = es.active_layer
        end
    end
    es.active_layer = es.active_layer + 1
    es.modified = true
    update_layers()
end

local
function move_layer_down()
    if es.active_layer == 1 or es.active_layer == 2 then
        return
    end
    es.layers[es.active_layer], es.layers[es.active_layer - 1] = 
        es.layers[es.active_layer - 1], es.layers[es.active_layer]
    for _, layer in ipairs(es.layers) do
        if layer.parent == es.active_layer then
            layer.parent = es.active_layer - 1
        elseif layer.parent == es.active_layer - 1 then
            layer.parent = es.active_layer
        end
    end
    es.active_layer = es.active_layer - 1
    es.modified = true
    update_layers()
end

local
function cycle_grid_rows()
    es.grid_rows = es.grid_rows % 7 + 1
    regen_grid()
    es.modified = true
    update_level_name()
end

local
function cycle_grid_cols()
    es.grid_cols = es.grid_cols % 7 + 1
    regen_grid()
    es.modified = true
    update_level_name()
end

local
function update_editor()
    local pressed = input_pressed()
    local pos = input_pos()
    local released = input_released()
    local down = input_down()

    -- timeline dragging
    if pressed and pos.y < timeline_top then
        es.timeline_press_x = pos.x
        es.timeline_x0 = (es.current_frame - 0.5) * frame_width
    elseif down and es.timeline_press_x then
        local delta_x = pos.x - es.timeline_press_x
        es.current_frame = math.clamp(math.ceil((es.timeline_x0 - delta_x) / frame_width), 1, timeline_frames)
        update_timeline()
    elseif released and es.timeline_press_x then
        es.timeline_press_x = nil
        es.timeline_x0 = nil
    elseif win:key_pressed"home" then
        goto_start()
    elseif win:key_pressed"end" then
        goto_end()
    end

    if win:key_pressed"l" then
        cycle_active_layer()
    end
    if win:key_pressed"p" then
        cycle_layer_parent()
    end
    if win:key_pressed"s" then
        cycle_layer_stencil()
    end
    if win:key_pressed"h" then
        cycle_layer_shader()
    end
    if win:key_pressed"u" then
        move_layer_up()
    end
    if win:key_pressed"d" then
        move_layer_down()
    end
    if win:key_pressed"r" then
        cycle_grid_rows()
    end
    if win:key_pressed"c" then
        cycle_grid_cols()
    end

    -- edit area touched
    if pressed and pos.y > edit_area_bottom and pos.y < edit_area_top then
        if es.mode == "translate" or es.mode == "rotate" or es.mode == "scale" then
            local layer = es.layers[es.active_layer]
            es.edit_start_pos = pos
            es.edit_keyframe0 = table.shallow_copy(get_or_create_keyframe(layer.keyframes, es.current_frame))
            es.timeline_quads:quad_color(es.frame_quads[es.current_frame], keyframe_color)
            es.modified = true
        elseif es.mode == "origin" then
            local layer = es.layers[es.active_layer]
            local keyframe = get_or_create_keyframe(layer.keyframes, es.current_frame)
            local origin = (math.inverse(layer.transform) * vec4(pos, 0, 1)).xy
            origin = vec2(snap(origin.x, snap_xy / keyframe.scale_x), snap(origin.y, snap_xy / keyframe.scale_y))
            layer.origin_cursor.position2d = origin
            keyframe.origin_x = origin.x
            keyframe.origin_y = origin.y
            es.timeline_quads:quad_color(es.frame_quads[es.current_frame], keyframe_color)
            es.modified = true
            update_layers()
        end
    elseif down and es.edit_start_pos then
        local layer = es.layers[es.active_layer]
        local keyframe = get_frame(layer.keyframes, es.current_frame) -- must be a keyframe due to call to get_or_create_keyframe above
        local mode = es.mode
        if mode == "translate" then
            local delta = pos - es.edit_start_pos
            if delta.x ~= 0 and not win:key_down"y" then
                keyframe.x = snap(es.edit_keyframe0.x + delta.x, snap_xy / keyframe.scale_x)
            end
            if delta.y ~= 0 and not win:key_down"x" then
                keyframe.y = snap(es.edit_keyframe0.y + delta.y, snap_xy / keyframe.scale_y)
            end
        elseif mode == "scale" then
            local delta = pos - es.edit_start_pos
            if win:key_down"x" then
                keyframe.scale_x = snap_scale(es.edit_keyframe0.scale_x + delta.x / 100)
            elseif win:key_down"y" then
                keyframe.scale_y = snap_scale(es.edit_keyframe0.scale_y + delta.x / 100)
            else
                keyframe.scale_x = snap_scale(es.edit_keyframe0.scale_x + delta.x / 100)
                keyframe.scale_y = snap_scale(es.edit_keyframe0.scale_y + delta.x / 100)
            end
        elseif mode == "rotate" then
            keyframe.angle = snap(es.edit_keyframe0.angle + (pos.x - es.edit_start_pos.x) * 0.01, snap_angle)
        end
        update_layers()
    elseif released and es.edit_start_pos then
        es.edit_start_pos = nil
        es.edit_keyframe0 = nil
    elseif down and (es.mode == "draw" or es.mode == "erase") and pos.y > edit_area_bottom and pos.y < edit_area_top then
        local row, column = find_active_layer_cell(pos)
        if row and column then
            local layer = es.layers[es.active_layer]
            if es.mode == "draw" then
                if layer.tiles[row][column] ~= es.draw_tile then
                    layer.tiles[row][column] = es.draw_tile
                    refresh_layer_mesh(layer)
                end
            else
                if layer.tiles[row][column] ~= 0 then
                    layer.tiles[row][column] = 0
                    refresh_layer_mesh(layer)
                end
            end
            es.modified = true
        end
    end

    es.clear_keyframe_button.disabled = not es.layers[es.active_layer].keyframes[es.current_frame]
    es.nudge_left_button.disabled = not es.layers[es.active_layer].keyframes[es.current_frame]
    es.nudge_right_button.disabled = not es.layers[es.active_layer].keyframes[es.current_frame]

    update_level_name()

    if win:key_pressed"escape" then
        win:close()
    end
end

local
function export_keyframes(ekeyframes)
    local keyframes = {}
    for f = 1, #ekeyframes do
        if ekeyframes[f] then
            local ekf = ekeyframes[f]
            local keyframe = {
                x = ekf.x,
                y = ekf.y,
                angle = ekf.angle,
                scale_x = ekf.scale_x,
                scale_y = ekf.scale_y,
                frame = f,
                origin_x = ekf.origin_x,
                origin_y = ekf.origin_y,
                ease = ekf.ease,
            }
            table.insert(keyframes, keyframe)
        end
    end
    return keyframes
end

local
function export_level()
    local layers = es.layers
    local level = {
        layers = {},
        name = es.level_name,
        version = es.level_version,
        grid_rows = es.grid_rows,
        grid_cols = es.grid_cols,
    }
    for l = 1, #layers do
        local layer = layers[l]
        local tiles
        if l > 1 then
            tiles = {}
            for i = 1, layer.height do
                tiles[i] = {}
                for j = 1, layer.width do
                    tiles[i][j] = layer.tiles[i][j]
                end
            end
        end
        local keyframes = export_keyframes(layer.keyframes)
        level.layers[l] = {
            tiles = tiles,
            keyframes = keyframes,
            parent = layer.parent,
            stencil = layer.stencil,
            shader = layer.shader,
        }
    end
    compute_level_collision_windows(level)
    set_bg_color(vec3(1), vec3(0))
    return level
end

local
function import_keyframes(keyframes)
    local ekeyframes = init_keyframes()
    for _, keyframe in ipairs(keyframes) do
        ekeyframes[keyframe.frame] = {
            x = keyframe.x,
            y = keyframe.y,
            angle = keyframe.angle,
            scale_x = keyframe.scale_x,
            scale_y = keyframe.scale_y,
            origin_x = keyframe.origin_x,
            origin_y = keyframe.origin_y,
            ease = keyframe.ease,
        }
    end
    return ekeyframes
end

local
function import_level(level)
    local layers = {}
    for l, layer in ipairs(level.layers) do
        local elayer
        if l == 1 then
            elayer = create_grid_layer(level.grid_rows, level.grid_cols)
        else
            local h, w = #layer.tiles, #layer.tiles[1]
            elayer = create_layer(w, h)
            for i = 1, h do
                for j = 1, w do
                    local tile = layer.tiles[i][j]
                    elayer.tiles[i][j] = tile
                end
            end
        end
        
        elayer.keyframes = import_keyframes(layer.keyframes)
        elayer.parent = layer.parent
        elayer.transform_dirty = true
        local curr_frame = get_frame(elayer.keyframes, 20)
        if l > 1 then
            elayer.origin_cursor.position2d = vec2(curr_frame.origin_x, curr_frame.origin_y)
            elayer.stencil = layer.stencil
            elayer.shader = layer.shader
            refresh_layer_mesh(elayer)
        end
        table.insert(layers, elayer)
    end
    es.layers = layers
    es.layer_group:remove_all()
    for _, layer in ipairs(layers) do
        es.layer_group:append(layer.node)
    end
    es.current_frame = 20
    es.active_layer = 2
    es.level_name = level.name
    es.level_version = level.version
    es.grid_rows = level.grid_rows
    es.grid_cols = level.grid_cols
    active_layer_change()
end

local socket

local
function delete_active_layer()
    if es.active_layer == 1 then
        return
    end
    es.modified = true
    local layer = es.layers[es.active_layer]
    es.layer_group:remove(layer.node)
    table.remove(es.layers, es.active_layer)
    for _, layer in ipairs(es.layers) do
        if layer.parent then
            if layer.parent == es.active_layer then
                layer.parent = nil
            elseif layer.parent > es.active_layer then
                layer.parent = layer.parent - 1
            end
        end
    end
    if #es.layers == 0 then
        local new_layer = create_layer(default_layer_w, default_layer_h)
        es.layer_group:append(new_layer.node)
        table.insert(es.layers, new_layer)
    elseif es.active_layer > #es.layers then
        es.active_layer = es.active_layer - 1
    end
    active_layer_change()
end

local
function create_new_layer()
    es.modified = true
    local new_layer = create_layer(default_layer_w, default_layer_h)
    es.layer_group:append(new_layer.node)
    table.insert(es.layers, new_layer)
    es.active_layer = #es.layers
    active_layer_change()
end

local
function all_blank_row(tiles, i)
    local num_cols = #tiles[1]
    for j = 1, num_cols do
        if tiles[i][j] > 0 then
            return false
        end
    end
    return true
end

local
function all_blank_col(tiles, j)
    local num_rows = #tiles
    for i = 1, num_rows do
        if tiles[i][j] > 0 then
            return false
        end
    end
    return true
end

local
function trim_level(level)
    level = table.deep_copy(level)
    for l = 2, #level.layers do
        local layer = level.layers[l]
        local tiles = layer.tiles
        local num_rows = #tiles
        local num_cols = #tiles[1]
        local row1, col1 = 1, 1
        local row2, col2 = num_rows, num_cols
        for j = 1, math.floor(num_cols/2) do
            if all_blank_col(tiles, j) and all_blank_col(tiles, num_cols - j + 1) then
                col1 = j + 1
                col2 = num_cols - j
            else
                break
            end
        end
        for i = 1, math.floor(num_rows/2) do
            if all_blank_row(tiles, i) and all_blank_row(tiles, num_rows - i + 1) then
                row1 = i + 1
                row2 = num_rows - i
            else
                break
            end
        end
        local new_tiles = {}
        for i = 1, num_rows do
            if i >= row1 and i <= row2 then
                local row = {}
                table.insert(new_tiles, row)
                for j = 1, num_cols do
                    if j >= col1 and j <= col2 then
                        table.insert(row, tiles[i][j])
                    end
                end
            end
        end
        layer.tiles = new_tiles             
    end
    return level
end

local
function do_play()
    clear_mesh_cache()
    local level = export_level()
    test_editor_level(level)
end

local
function next_name()
    local name = save_index.next_name
    local chars = {}
    local a = string.byte("A")
    local z = string.byte("Z")
    for i = 1, #name do
        chars[#name - i + 1] = name:byte(i)
    end
    local i = 1
    while true do
        chars[i] = chars[i] + 1
        if chars[i] > z then
            chars[i] = a
            i = i + 1
            if i > #name then
                error("reached name limit")
            end
        else
            break
        end
    end
    local new_name = ""
    for i = 1, #name do
        new_name = string.char(chars[i])..new_name
    end
    save_index.next_name = new_name
    return name
end

local
function save_level(level_history)
    local latest = level_history[#level_history]
    if latest then
        save_json("src/levels/"..latest.name..".json", trim_level(latest))
        save_json("editor_data/"..latest.name..".json", level_history)
    end
end

local
function load_level_history(name)
    return load_json("editor_data/"..name..".json")
end

local
function inc_version_and_save()
    if not es.modified then
        return
    end
    es.modified = nil
    local level = export_level()
    local level_history = load_level_history(level.name) or {}
    level.version = #level_history + 1
    level_history[level.version] = level
    save_index.levels[es.level_name] = true
    save_index.last_open = level.name
    save_save_index()
    save_level(level_history)
    es.level_version = level.version
end

local
function new_level()
    es.modified = nil
    es.grid_rows = 3
    es.grid_cols = 3
    es.layers = {create_grid_layer(es.grid_rows, es.grid_cols), create_layer(default_layer_w, default_layer_h)}
    es.active_layer = 2
    es.current_frame = 20
    es.layer_group:remove_all()
    for _, layer in ipairs(es.layers) do 
        es.layer_group:append(layer.node)
    end
    es.level_name = next_name()
    es.level_version = 0
    active_layer_change()
end

local
function load_level(name, version)
    local history = load_level_history(name)
    if not history then
        return
    end
    version = version or #history
    if version > #history or version < 1 then
        return
    end
    local level = history[version]
    import_level(level)
    save_index.last_open = name
    save_save_index()
    es.modified = nil
end

local
function open_level()
    local overlay = am.rect(win.left, win.bottom, win.right, win.top, vec4(0, 0, 0, 0.5))
    local columns = 10
    local level_names = table.keys(save_index.levels)
    table.sort(level_names)
    local num_levels = #level_names
    local rows = math.ceil(num_levels / columns)
    local cell_w = win.width / columns
    local cell_h = 50
    local y = 0
    local level = 1
    local scrollarea = am.translate(0, win.height/2)
    for row = 1, rows do
        local x = win.left
        for column = 1, columns do
            if level <= num_levels then
                local color = (row + column) % 2 == 0 and vec4(0.5, 0.5, 0.5, 1) or vec4(0.4, 0.4, 0.4, 1)
                local node = am.translate(x + cell_w/2, y - cell_h/2) ^ {
                    am.rect(-cell_w/2, -cell_h/2, cell_w/2, cell_h/2, color):tag("level"..level),
                    am.scale(2) ^ am.text(level_names[level])
                }
                scrollarea:append(node)
                x = x + cell_w   
                level = level + 1
            end
        end
        y = y - cell_h
    end

    local drag_pos
    local start_pos
    scrollarea:action(function()
        if input_pressed() then
            drag_pos = input_pos()
            start_pos = scrollarea.position2d
        elseif input_released() and drag_pos then
            if math.distance(drag_pos, input_pos()) < 10 then
                local pos = input_pos()
                local column = math.clamp(math.ceil(((pos.x + win.width/2) / win.width) * columns), 1, columns)
                local row = math.ceil(- (pos.y - scrollarea.position2d.y) / cell_h)
                local level = (row - 1) * columns + column
                if level >= 1 and level <= num_levels then
                    scrollarea("level"..level).color = vec4(0.8, 0.8, 0.8, 1)
                    scrollarea:action(am.series{
                        am.delay(0.1),
                        function()
                            root_scene:remove(scrollarea)
                            root_scene:remove(overlay)
                            es.editor_node.paused = false
                            load_level(level_names[level])
                            return true
                        end
                    })
                else
                    root_scene:remove(scrollarea)
                    root_scene:remove(overlay)
                    es.editor_node.paused = false
                    return true
                end
            end
            drag_pos = nil
            start_pos = nil
        elseif input_down() and drag_pos then
            --scrollarea.position2d = vec2(start_pos.x,
            --    math.clamp(start_pos.y + 2 * (input_pos().y - drag_pos.y), win.height/2, cell_h*rows-win.height/2))
        end
    end)

    es.editor_node.paused = true
    root_scene:append(overlay)
    root_scene:append(scrollarea)
end

local
function clear_current_keyframe()
    es.modified = true
    local layer = es.layers[es.active_layer]
    layer.keyframes[es.current_frame] = false
    es.timeline_quads:quad_color(es.frame_quads[es.current_frame], get_timeline_notch_color(es.current_frame))
    update_layers()
end

local
function nudge_current_keyframe(dir)
    es.modified = true
    local curr = es.current_frame
    local layer = es.layers[es.active_layer]
    if dir + curr < 1 or dir + curr > timeline_frames then
        return
    end
    if not layer.keyframes[curr] or layer.keyframes[curr + dir] then
        return
    end
    layer.keyframes[curr + dir] = layer.keyframes[curr]
    layer.keyframes[curr] = false
    es.current_frame = curr + dir
    active_layer_change()
end

local
function insert_frame()
    es.modified = true
    local curr = es.current_frame
    for _, layer in ipairs(es.layers) do
        for k = timeline_frames, curr+1, -1 do
            layer.keyframes[k] = layer.keyframes[k-1]
        end
        layer.keyframes[curr] = false
    end
    active_layer_change()
end

local
function delete_frame()
    es.modified = true
    local curr = es.current_frame
    for _, layer in ipairs(es.layers) do
        for k = curr, timeline_frames-1 do
            layer.keyframes[k] = layer.keyframes[k+1]
        end
        layer.keyframes[timeline_frames] = false
    end
    active_layer_change()
end

local
function set_easing(ease)
    es.ease = ease
    local keyframes = es.layers[es.active_layer].keyframes
    local keyframe = get_prev_keyframe(keyframes, es.current_frame)
    if keyframe then
        keyframe.ease = ease
        es.modified = true
        active_layer_change()
    end
end

local
function create_device_bounds_rect(aspect, phone)
    local w, h
    if phone then
        w = 280
        h = 280
    else
        w = 340
        h = 340
    end
    if aspect > 1 then
        w = h * aspect
    else
        h = w / aspect
    end
    local x1 = -w/2
    local x2 = w/2
    local y1 = -h/2
    local y2 = h/2
    local thickness = 1
    local color = vec4(1, 0, 1, 1)
    return am.group{
        am.line(vec2(x1, y1), vec2(x2, y1), thickness, color),
        am.line(vec2(x1, y1), vec2(x1, y2), thickness, color),
        am.line(vec2(x2, y1), vec2(x2, y2), thickness, color),
        am.line(vec2(x2, y2), vec2(x1, y2), thickness, color),
    }
end

local
function create_device_bounds()
    return am.group{
        create_device_bounds_rect(19.5/9),
        create_device_bounds_rect(9/19.5, true),
        create_device_bounds_rect(9/16, true),
    }
end

function set_editor_player_recording(player_recording)
    es.player_recording = player_recording
end

local
function create_editor()
    local timeline_quads, frame_quads = create_timeline_quads()
    local current_frame_line = am.line(vec2(0, win.bottom), vec2(0, timeline_top + 16), 2)
    local current_frame_text = am.text("XXX", "left", "bottom")
    local timeline_translate = am.translate(0, 0) ^ timeline_quads
    local buttons_group = am.group()

    local clear_keyframe_button = create_button(editor_sprites.clear_keyframe, win.right - 45, timeline_top, function()
        clear_current_keyframe()
    end)
    buttons_group:append(clear_keyframe_button)
    local nudge_left_button = create_button(editor_sprites.nudge_left, win.right - 135, timeline_top, function()
        nudge_current_keyframe(-1)
    end)
    buttons_group:append(nudge_left_button)
    local nudge_right_button = create_button(editor_sprites.nudge_right, win.right - 90, timeline_top, function()
        nudge_current_keyframe(1)
    end)
    buttons_group:append(nudge_right_button)

    local insert_frame_button = create_button(editor_sprites.insert_frame, win.right - 225, timeline_top, function()
        insert_frame()
    end)
    buttons_group:append(insert_frame_button)
    local delete_frame_button = create_button(editor_sprites.delete_frame, win.right - 270, timeline_top, function()
        delete_frame()
    end)
    buttons_group:append(delete_frame_button)

    local ease_buttons = {}
    for i = 1, #layer_easings do
        local ease_button = create_button(editor_sprites["ease"..i], win.right - 400 - (i-1)*45, timeline_top, function()
            set_easing(i)
        end, "ease")
        buttons_group:append(ease_button)
        ease_buttons[i] = ease_button
    end

    local play_button = create_button(editor_sprites.play, win.left, timeline_top, function()
        do_play()
    end)
    buttons_group:append(play_button)
    local origin_button = create_button(editor_sprites.origin, win.right - 45 * 4, win.top - 45, function()
        es.mode = "origin"
    end, "mode")
    buttons_group:append(origin_button)
    local translate_button = create_button(editor_sprites.translate, win.right - 45 * 3, win.top - 45, function()
        es.mode = "translate"
    end, "mode")
    buttons_group:append(translate_button)
    local scale_button = create_button(editor_sprites.scale, win.right - 45 * 2, win.top - 45, function()
        es.mode = "scale"
    end, "mode")
    buttons_group:append(scale_button)
    local rotate_button = create_button(editor_sprites.rotate, win.right - 45 * 1, win.top - 45, function()
        es.mode = "rotate"
    end, "mode")
    buttons_group:append(rotate_button)
    local draw1_button = create_button(editor_sprites.draw1, win.right - 45 * 7, win.top - 90, function()
        es.mode = "draw"
        es.draw_tile = 1
    end, "mode")
    buttons_group:append(draw1_button)
    for d = 2, 13 do
        local col = math.floor((d - 2) / 2)
        local row = (d - 2) % 2
        if col % 2 == 0 then
            row = 1 - row
        end
        local draw_button = create_button(editor_sprites["draw"..d], win.right - 45 * 8 - 45 * col, win.top - (row + 1) * 45, function()
            es.mode = "draw"
            es.draw_tile = d
        end, "mode")
        buttons_group:append(draw_button)
    end
    local erase_button = create_button(editor_sprites.erase, win.right - 45 * 7, win.top - 45, function()
        es.mode = "erase"
    end, "mode")
    buttons_group:append(erase_button)
    local delete_layer_button = create_button(editor_sprites.delete_layer, win.left, win.top - 90, function()
        delete_active_layer()
    end)
    buttons_group:append(delete_layer_button)
    local new_layer_button = create_button(editor_sprites.new_layer, win.left + 45, win.top - 90, function()
        create_new_layer()
    end)
    buttons_group:append(new_layer_button)
    local undo_button = create_button(editor_sprites.undo, win.left + 135, win.top - 90, function()
        if es.modified then
            load_level(es.level_name, es.level_version)
        else
            load_level(es.level_name, es.level_version - 1)
        end
        es.modified = true
    end)
    buttons_group:append(undo_button)
    local redo_button = create_button(editor_sprites.redo, win.left + 180, win.top - 90, function()
        load_level(es.level_name, es.level_version + 1)
        es.modified = true
    end)
    buttons_group:append(redo_button)
    local next_layer_button = create_button(editor_sprites.next_layer, win.left + 90, win.top - 90, function()
        cycle_active_layer()
    end)
    buttons_group:append(next_layer_button)
    local save_button = create_button(editor_sprites.save, win.left + 45, win.top - 45, function()
        inc_version_and_save()
    end)
    buttons_group:append(save_button)
    local open_button = create_button(editor_sprites.load, win.left + 90, win.top - 45, function()
        open_level()
    end)
    buttons_group:append(open_button)
    local new_button = create_button(editor_sprites.new, win.left + 135, win.top - 45, function()
        new_level()
    end)
    buttons_group:append(new_button)
    local level_name_text = am.text("XXXXvXXXX", "left", "top")
    local level_name_node = am.translate(win.left, win.top - 90) ^ level_name_text
    local layer_info_node = am.translate(win.left, win.top - 108) ^ am.text("", "left", "top")
    local layer_group = am.group()
    local device_bounds = create_device_bounds()
    local player_pos_node = am.translate(0, 0) ^ am.sprite(editor_sprites.player_pos)
    local editor_node =
        am.use_program(am.shaders.colors2d)
        ^ {
            layer_group
        ,
            device_bounds
        ,
            player_pos_node
        ,
            level_name_node
        ,
            layer_info_node
        ,
            timeline_translate
        ,
            am.translate(4, timeline_top)
            ^ current_frame_text
        ,
            current_frame_line
        ,
            buttons_group
        }

    table.clear(es)
    es.editor_node = editor_node
    es.layer_info_node = layer_info_node
    es.layer_group = layer_group
    es.clear_keyframe_button = clear_keyframe_button
    es.nudge_left_button = nudge_left_button
    es.nudge_right_button = nudge_right_button
    es.play_button = play_button
    es.translate_button = translate_button
    es.scale_button = scale_button
    es.rotate_button = rotate_button
    es.erase_button = erase_button
    es.current_frame_text = current_frame_text
    es.frame_quads = frame_quads
    es.timeline_quads = timeline_quads
    es.timeline_translate = timeline_translate
    es.buttons_group = buttons_group
    es.player_pos_node = player_pos_node
    es.level_name_text = level_name_text
    es.ease_buttons = ease_buttons
    es.ease = 1

    button_toggle_groups.ease = es.ease_buttons[es.ease]
    button_toggle_groups.ease.toggled = true
    if arg[1] and arg[1]:match("[A-Z][A-Z][A-Z]") then
        load_level(arg[1])
        table.remove(arg, 1)
    elseif save_index.last_open then
        load_level(save_index.last_open)
    else
        new_level()
    end
    editor_node:action("editor_action", update_editor)
end

function resume_editing()
    set_bg_color(vec3(1), vec3(0))
    root_scene:remove_all()
    root_scene:append(es.editor_node)
end

create_editor()
resume_editing()

noglobals()
