require "window"
require "geometry"
require "level"
require "player"
require "background"
require "level_select"
require "level_visuals"
require "explosion"
require "level_order"
require "audio"
require "ending"
require "util"
require "leaderboards"
require "achievements"
require "tutorial"
local sprites = require "sprites"
local palettes = require "palettes"

-- for screenshot mode
local manual_t = 0.2258

local
function init_stage_stats()
    return {
        secs = 0,
        retries = 0,
        jumps = 0,
    }
end

local
function record_best_time(gs)
    if not global_state.best_times then
        global_state.best_times = {}
        for i = 1, num_stages do
            global_state.best_times[i] = false
        end
    end
    if not global_state.best_times[gs.stage_num] or global_state.best_times[gs.stage_num] > gs.stage_stats.secs then
        global_state.best_times[gs.stage_num] = gs.stage_stats.secs
        save_state()
    end
end

local
function compute_play_area(gs)
    local aspect = win.width / win.height
    local w, h
    if aspect > 1 then
        h = settings.min_screen_height
        w = aspect * h
    else
        w = settings.min_screen_width
        h = w / aspect
    end
    gs.play_area_width = w
    gs.play_area_height = h
end

function init_game_state(level, level_num, mode, stats)
    local is_compound_level = not not level.sublevels
    local is_flipped = is_phone and is_level_flipped_on_phone[level.name]
        or not is_phone and is_level_flipped_on_non_phone[level.name]
    local gs = {
        main_scene = is_flipped and am.rotate(math.pi*0.5) or am.group(),
        real_time = 0,
        is_gameover = false,
        level_num = level_num,
        mode = mode,
        stage_stats = stats or init_stage_stats(),
        time_scale = settings.time_scale,
        stage_num = get_level_stage(level_num),
        rand = am.rand(os.time()),
        is_flipped = is_flipped,
    }
    if mode == "campaign" and global_state.game_speed then
        gs.time_scale = gs.time_scale * (global_state.game_speed * 0.01)
    end
    if is_compound_level then
        gs.compound_level = level
        gs.level = level.sublevels[1].level
        gs.compound_level_index = 1
    else
        gs.level = level
    end
    gs.palette = get_level_palette(gs.level_num)
    compute_play_area(gs)
    init_background(gs)
    if is_compound_level then
        init_vortex_as_background(gs)
    end
    gs.animated_grid = #gs.level.layers[1].keyframes > 0
    init_grid(gs)
    init_tokens(gs)
    init_explosion(gs)
    init_level(gs)
    init_player(gs)
    update_level(gs)
    update_level_node_transforms(gs)
    update_grid(gs)
    update_tokens(gs)
    update_player_graphics(gs)
    return gs
end

local
function set_ortho_P(gs)
    local w, h = gs.play_area_width, gs.play_area_height
    gs.proj_mat = math.ortho(-w/2, w/2, -h/2, h/2, -300, 500)
    gs.top_bind.P = gs.proj_mat
end

local
function handle_resize(gs)
    if win:resized() then
        compute_play_area(gs)
        set_ortho_P(gs)
        if is_mobile then
            gs.close_button.x = win.safe_left + 25
            gs.close_button.y = win.safe_top - 20
        end
    end
end

local
function handle_escape(gs)
    if gs.mode == "stage_speedrun" then
        show_speedrun_menu()
    else
        show_title()
    end
end

local
function save_progress(gs)
    if is_editor then
        return
    end
    local slot
    if gs.mode == "campaign" then
        slot = "saved_game"
    elseif gs.mode == "stage_speedrun" then
        slot = "speedrun_save"..gs.stage_num
    end
    if slot then
        if not global_state[slot] then
            global_state[slot] = {
                level = gs.level_num,
            }
        else
            global_state[slot].level = gs.level_num
        end
        global_state[slot].stage_stats = gs.stage_stats
    end
    save_state()
end

local
function update_colors(gs)
    if global_state.less_flashing then
        return
    end
    local palette1 = gs.palette
    local palette2 = palettes[gs.stage_num + 4]
    local r
    if gs.compound_level then
        r = (-math.cos(gs.real_time * 2) + 1) * 0.5
    else
        if gs.tokens_left == 1 then
            r = 0
        else
            r = 1 - (gs.tokens_left-2) / (gs.total_tokens-2)
        end
    end
    if r == gs.prev_color_cycle_r then
        return
    end
    gs.prev_color_cycle_r = r
    local bg_i = gs.compound_level and 9 or 4
    local grid_i = gs.compound_level and 10 or 3
    local star_color = math.mix(palette1[5], palette2[5], r) 
    local bg_color = math.mix(palette1[bg_i], palette2[bg_i], r)
    if not gs.compound_level then
        set_bg_color(star_color, bg_color)
    end

    local grid_color = math.mix(palette1[grid_i], palette2[grid_i], r)
    local grid_color4 = vec4(grid_color, 1)
    for _, line in ipairs(gs.grid_overlay:all"line") do
        line.color = grid_color4
    end
    if not gs.compound_level then
        gs.tokens.node.color = grid_color
        gs.tokens.node.shadow_color = bg_color
        gs.token_collect_particles.start_color = grid_color4
        gs.token_collect_particles.end_color = grid_color4
    end

    set_vortex_color(bg_color, grid_color)

    for _, layer in ipairs(gs.layer_states) do
        if layer.node then
            local colors = layer_shaders[layer.shader].colors
            local main_color = math.mix(palette1[colors[1]], palette2[colors[1]], r)
            local detail_color1 = math.mix(palette1[colors[2]], palette2[colors[2]], r)
            local detail_color2 = math.mix(palette1[colors[3]], palette2[colors[3]], r)
            local bind = layer.node"bind"
            bind.main_color = main_color
            bind.detail_color1 = detail_color1
            bind.detail_color2 = detail_color2
        end
    end
end

local
function save_end_campaign_stage(gs)
    if gs.mode == "campaign" then
        if gs.level_num >= num_levels then
            global_state.saved_game = nil
        else
            global_state.saved_game.level = gs.level_num + 1
            global_state.saved_game.stage_stats = init_stage_stats()
        end
        save_state()
    elseif gs.mode == "stage_speedrun" then
        global_state["speedrun_save"..gs.stage_num] = nil
        save_state()
    end
end

local
function start_game(level_num, level, mode, stats, no_reset_input_timer)
    if not no_reset_input_timer then
        last_movement_input_time = am.frame_time
    end
    if not level and level_num then
        local lvl = level_order[level_num]
        if is_mobile and mobile_replacement_levels[lvl] then
            lvl = mobile_replacement_levels[lvl]
        end
        if type(lvl) == "string" then
            level = am.parse_json(am.load_string("levels/"..lvl..".json"))
        else
            -- timed level
            level = {
                name = lvl.name,
                sublevels = {},
            }
            for _, name in ipairs(lvl.sublevels) do
                if is_mobile and mobile_replacement_levels[name] then
                    name = mobile_replacement_levels[name]
                end
                table.insert(level.sublevels, {
                    name = name,
                    level = am.parse_json(am.load_string("levels/"..name..".json")),
                })
            end
        end
    end

    print(level.name.."("..level_num..")")

    local gs = init_game_state(level, level_num, mode, stats)
    if is_editor then
        gs.input_recording = {{time = 0, x = gs.player_pos.x, y = gs.player_pos.y}}
    end

    save_progress(gs)

    local
    function gameover_action(gs)
        if gs.token_collect_particles then
            gs.token_collect_particles.hidden = true
        end
        gs.teleport_particles.hidden = true
        if gs.is_stage_cleared then
            if is_editor then
                return function()
                    resume_editing(gs.input_recording)
                    return true
                end
            end
            local next_level = gs.level_num + 1
            play_audio_warp()
            if not gs.compound_level then
                vortex_zoom(gs, type(level_order[next_level]) == "table")
            end
            gs.grid_overlay.hidden = true
            set_star_speed(800)
            gs.layers_group.hidden = true
            if gs.tutorial_node then
                gs.main_scene:remove(gs.tutorial_node)
            end
            gs.player_node"scale":cancel("jump") -- cancel jump scale tween
            gs.player_node"scale".scale2d = vec2(0.15)
            -- continue animating player
            gs.player_node:action(function()
                update_player_graphics(gs)
            end)
            if global_state.less_flashing then
                fade_player_out(gs)
            end
            if gs.compound_level then
                gs.main_scene:action("colors", function() 
                    gs.real_time = gs.real_time + am.delta_time
                    update_colors(gs)
                end)
                gs.pause_timer = true
                if gs.timer_node then
                    gs.timer_node.hidden = true
                end
                vortex_speedup_background()
                if not global_state.asked_for_review and am.ios_request_review then
                    am.ios_request_review()
                    global_state.asked_for_review = true
                end
                save_end_campaign_stage(gs)
                submit_score(gs)
                record_best_time(gs)
                local action_seq = {
                    am.delay(0.1),
                    am.tween(gs.player_node"scale", 0.3, {scale2d = vec2(4)}, am.ease.hyperbola),
                    function()
                        start_end_stage_music()
                        gs.player_node.hidden = true
                        return true
                    end,
                    am.delay(0.5),
                    function()
                        local text1 = strings.quadrant_cleared--.." ["..get_level_stage(gs.level_num).."/"..num_stages.."]"
                        local text2a = strings.time
                        local text2b = format_secs(gs.stage_stats.secs)
                        local text3a = strings.retries
                        local text3b = tostring(gs.stage_stats.retries)
                        local text4a = strings.jumps
                        local text4b = tostring(gs.stage_stats.jumps)
                        local color = vec4(gs.palette[6], 1)
                        local line1 = am.scale(0.5) ^ am.text(font, text1, color)
                        local align = 400
                        local line2 = am.scale(0.5) ^ {
                            am.translate(-align, 0) ^ am.text(font, text2a, color, "left"),
                            am.translate(align, 0) ^ am.text(font, text2b, color, "right"),
                        }
                        local line3 = am.scale(0.5) ^ {
                            am.translate(-align, 0) ^ am.text(font, text3a, color, "left"),
                            am.translate(align, 0) ^ am.text(font, text3b, color, "right"),
                        }
                        local line4 = am.scale(0.5) ^ {
                            am.translate(-align, 0) ^ am.text(font, text4a, color, "left"),
                            am.translate(align, 0) ^ am.text(font, text4b, color, "right"),
                        }
                        line1.hidden = true
                        line2.hidden = true
                        line3.hidden = true
                        line4.hidden = true
                        local scl = is_phone and 0.6 or 1

                        local menu
                        local menu_items = {
                            {
                                text = function() return strings.continue end,
                                mobile_big = true,
                                action = function()
                                    gs.main_scene:cancel("colors")
                                    if gs.mode == "campaign" then
                                        gs.top_scene:remove(menu)
                                        gs.top_scene:action(am.series{
                                            function()
                                                start_music(gs.level_num % num_levels + 1)
                                                gs.top_scene:remove(gs.stage_clear_message)
                                                show_background()
                                                play_audio_warp()
                                                vortex_zoom_background()
                                                return true
                                            end,
                                            am.delay(1.0),
                                            function()
                                                if next_level > num_levels then
                                                    show_ending()
                                                else
                                                    start_game(next_level, nil, gs.mode, init_stage_stats())
                                                end
                                                return true
                                            end,
                                        })
                                    elseif gs.mode == "stage_speedrun" then
                                        show_speedrun_menu()
                                    end
                                    return true
                                end,
                            },
                        }
                        if gs.mode == "campaign" then
                            table.insert(menu_items, {
                                text = function() return strings.back end,
                                action = function()
                                    show_title()
                                    return true
                                end
                            })
                        end
                        menu = make_menu(menu_items, win.safe_bottom + 100, gs.palette, 1)

                        local node = am.translate(0, 0) 
                            ^ am.scale(scl)
                            ^ {
                                am.translate(0, 180) ^ line1,
                                am.translate(0, 70) ^ line2,
                                am.translate(0, 20) ^ line3,
                                am.translate(0, -30) ^ line4,
                            }
                        node:action(coroutine.create(function()
                            am.wait(am.delay(0.5))
                            line1.hidden = false
                            am.wait(am.delay(0.5))
                            line2.hidden = false
                            am.wait(am.delay(0.2))
                            line3.hidden = false
                            am.wait(am.delay(0.2))
                            line4.hidden = false
                            am.wait(am.delay(0.5))
                            gs.top_scene:append(menu)
                        end))
                        gs.top_scene:append(node)
                        gs.stage_clear_message = node
                        return true
                    end
                }
                return am.series(action_seq)
            else
                return 
                    am.parallel{
                        am.series{
                            am.delay(0.1),
                            am.tween(gs.player_node"scale", 0.3, {scale2d = vec2(4)}, am.ease.hyperbola),
                            function()
                                gs.player_node.hidden = true
                                return true
                            end
                        },
                        am.series{
                            am.delay(1.0),
                            function()
                                if next_level > num_levels then
                                    show_ending()
                                else
                                    start_game(next_level, nil, gs.mode, gs.stage_stats)
                                end
                                return true
                            end
                        },
                    }
            end
        else
            -- player died
            gs.stage_stats.retries = gs.stage_stats.retries + 1
            play_audio_gameover()
            set_star_speed(0)
            make_explosion(gs)
            return am.parallel{
                am.series{
                    am.delay(0.75),
                    function()
                        start_game(gs.level_num, gs.compound_level or gs.level, gs.mode, gs.stage_stats, true)
                        return true
                    end
                },
                function()
                    if global_state.less_flashing then
                        for _, bind in ipairs(gs.player_color_binds) do
                            bind.color = gs.palette[6]
                        end
                        return true
                    else
                        if math.fract(am.frame_time * 7) < 0.5 then
                            for _, bind in ipairs(gs.player_color_binds) do
                                bind.color = gs.palette[6]
                            end
                        else
                            for _, bind in ipairs(gs.player_color_binds) do
                                bind.color = gs.palette[4]
                            end
                        end
                    end
                end
            }
        end
    end
    local
    function main_action()
        if not is_editor and gs.mode == "campaign" and am.frame_time - last_movement_input_time > settings.no_input_reset_secs then
            show_title()
            return true
        end
        gs.real_time = gs.real_time + am.delta_time
        if handle_player_jump(gs) then
            gs.stage_stats.jumps = gs.stage_stats.jumps + 1
        end
        local next_keyframe_time = gs.keyframe_time + am.delta_time * gs.time_scale
        local collide_t = check_player_collision_window(gs, gs.keyframe_time, next_keyframe_time)
        if collide_t then
            gs.keyframe_time = collide_t
            gs.is_gameover = true
        else
            gs.keyframe_time = next_keyframe_time
        end
        update_level(gs)
        update_player_graphics(gs)
        if not gs.is_gameover then
            update_tokens(gs)
        end
        update_level_node_transforms(gs)
        update_grid(gs)
        if screenshot_mode then
            if win:key_down"t" then
                manual_t = math.fract(manual_t + am.delta_time * 0.1)
            end
            gs.top_bind.t = manual_t
        else
            gs.top_bind.t = gs.real_time
        end
        if gs.is_gameover or gs.is_stage_cleared then
            check_achievements(gs)
            gs.main_scene:action("main", gameover_action(gs))
            return true
        end
        update_colors(gs)
    end
    gs.main_scene:action("main", main_action)
    gs.top_scene = 
        am.group()
        ^ am.bind{
            P = mat4(1),
            color_offset = vec3(0),
            t = 0,
            explosion_rad = 0,
            explosion_pos = vec2(0),
            explosion_color = gs.palette[4],
            light_dir = math.normalize(vec3(1, 0, 1)),
        }
        ^ gs.main_scene
    gs.top_bind = gs.top_scene"bind"
    set_ortho_P(gs)

    if mode == "stage_speedrun" then
        local timer_color = vec4(gs.palette[6], 1)
        if global_state.show_speedrun_timer then
            local timer_node = am.translate(win.safe_right - 10, win.safe_top - 10) 
                ^ am.scale(0.35) 
                ^ am.text(timer_font, format_secs(0, 1), timer_color, "right", "top")
            gs.top_scene:append(timer_node)
            gs.timer_node = timer_node
        end
    end

    if not capture_mode and not screenshot_mode then
        local name_color = vec4(gs.palette[6], 1)
        local name_node = am.translate(win.safe_right - 10, win.safe_bottom + 10)
            ^ am.scale(0.3)
            ^ am.text(font, gs.stage_num.." - "..(gs.level_num - get_stage_first_level(gs.stage_num) + 1), name_color, "right", "bottom")
        gs.top_scene:append(name_node)
        name_node:action(am.series{
            am.delay(0.5),
            am.tween(name_node"text", 0.5, {color = name_color{a = 0}}),
            function()
                gs.top_scene:remove(name_node)
                return true
            end
        })
    end

    if is_mobile then
        local color = vec4(gs.palette[6], 1)
        local close_button = am.translate(win.safe_left + 25, win.safe_top - 20)
            ^ am.scale(0.25) ^ am.sprite(sprites.cross, color)
        gs.top_scene:append(close_button)
        gs.close_button = close_button
    end

    gs.top_scene:action(function() 
        handle_resize(gs)
        if esc_pressed() then
            if is_editor then
                resume_editing()
            else
                gs.main_scene:cancel"main"
                handle_escape(gs)
                return true
            end
        end
        if is_mobile and input_pressed() then
            local pos = input_pos()
            if pos.x < win.safe_left + 50 and pos.y > win.safe_top - 50 then
                handle_escape(gs)
                gs.main_scene:cancel"main"
                play_audio_gameover()
                return true
            end
        end
        if not gs.pause_timer then
            gs.stage_stats.secs = gs.stage_stats.secs + am.delta_time
            if gs.timer_node then
                gs.timer_node"text".text = format_secs(gs.stage_stats.secs, 1)
            end
        end
    end)
    root_scene:remove_all()
    root_scene:append(gs.top_scene)
    main_action()

    if gs.mode == "campaign" then
        if level_num == 1 then
            show_tutorial(gs)
        elseif is_mobile and gs.level.name == "ACS" then
            show_tutorial(gs)
        end
    end

    start_music(level_num)
end

local
function start_transition(l, f)
    start_music(l)
    local bg = get_bg() 
    local palette = get_level_palette(l)
    local vortex = init_vortex(palette)
    vortex_zoom({palette = get_level_palette(l)}, false)
    set_star_speed(800)
    set_bg_color(palette[5], palette[4])
    show_background()
    play_audio_warp()
    root_scene:remove_all()
    root_scene:append(am.group{bg, vortex})
    root_scene:action(am.series{
        am.delay(1),
        function()
            f()
            return true
        end,
    })
end

function start_new_campaign()
    start_transition(1, function()
        start_game(1, nil, "campaign", init_stage_stats())
    end)
end

function test_editor_level(level)
    start_game(2, level, "campaign", init_stage_stats())
end

function test_level(l)
    start_game(l, nil, "campaign", init_stage_stats())
end

function continue_campaign()
    local l = global_state.saved_game.level
    start_transition(l, function()
        start_game(l, nil, "campaign", global_state.saved_game.stage_stats)
    end)
end

function start_stage_speedrun(stage)
    local l = get_stage_first_level(stage)
    start_transition(l, function()
        start_game(l, nil, "stage_speedrun", init_stage_stats())
    end)
end

function continue_stage_speedrun(stage)
    local save = global_state["speedrun_save"..stage]
    if not save then
        start_stage_speedrun(stage)
    else
        local l = save.level
        start_transition(l, function()
            start_game(l, nil, "stage_speedrun", save.stage_stats)
        end)
    end
end
