require "window"
require "geometry"
require "level"
require "player"
require "background"
require "level_select"
require "level_visuals"
require "explosion"
require "level_order"
require "audio"
require "ending"
require "util"
require "leaderboards"
require "achievements"
require "tutorial"
require "gameover_inf"
local sprites = require "sprites"
local palettes = require "palettes"

-- for screenshot mode
local manual_t = 0.2258

local score_scale = vec2(0.4)

local
function init_stage_stats()
    return {
        secs = 0,
        retries = 0,
        jumps = 0,
    }
end

local
function compute_play_area(gs)
    local aspect = win.width / win.height
    local w, h
    if aspect > 1 then
        h = settings.min_screen_height
        w = aspect * h
    else
        w = settings.min_screen_width
        h = w / aspect
    end
    gs.play_area_width = w
    gs.play_area_height = h
end

local inf_level_catalog

local
function pick_next_pattern_level(gs)
    if not inf_level_catalog then
        inf_level_catalog = {}
        for p, phase in ipairs(infinite_levels) do
            inf_level_catalog[p] = {}
            for l, lvl in ipairs(infinite_levels[p]) do
                if is_mobile and mobile_replacement_levels[lvl] then
                    lvl = mobile_replacement_levels[lvl]
                end
                inf_level_catalog[p][l] = am.parse_json(am.load_string("levels/"..lvl..".json"))
                inf_level_catalog[p][l].inf_stage = (p - 1) % num_stages + 1
                inf_level_catalog[p][l].palette_id = (l - 1) % num_stages + 1
            end
        end
    end
    if not gs.level_bucket then
        gs.level_bucket = {}
    end
    if #gs.level_bucket == 0 then
        -- refill bucket
        gs.inf_phase = gs.inf_phase + 1
        if gs.inf_phase <= (#infinite_levels - 1) then
            for _, level in ipairs(inf_level_catalog[gs.inf_phase]) do
                for i = 1, 2 do
                    table.insert(gs.level_bucket, level)
                end
            end
        elseif gs.inf_phase == #infinite_levels then
            table.insert(gs.level_bucket, inf_level_catalog[gs.inf_phase][1])
        else
            for i = 1, 10 do
                local phase = gs.rand(#infinite_levels - 1)
                local l = gs.rand(#infinite_levels[phase])
                table.insert(gs.level_bucket, inf_level_catalog[phase][l])
            end
        end
        table.shuffle(gs.level_bucket, gs.rand)
    end
    return table.remove(gs.level_bucket)
end

local
function start_next_pattern(gs)
    local level = pick_next_pattern_level(gs)
    print(level.name)
    local pattern = {
        level = level,
        main_scene = am.group(),
        is_gameover = false,
        mode = "infinite",
        time_scale = gs.time_scale,
        palette = palettes[level.palette_id],
        play_area_width = gs.play_area_width,
        play_area_height = gs.play_area_height,
        stage_num = level.inf_stage,
        level_num = gs.level_num,
        player_transform = gs.player_transform,
        is_flipped = gs.is_flipped,
    }
    init_level(pattern)
    update_level(pattern)
    update_level_node_transforms(pattern)
    gs.level_scene:append(pattern.main_scene)
    table.insert(gs.patterns, pattern)
end

local
function update_patterns(gs)
    -- work out when the currently playing patterns will end
    local end_t = 0
    for i = #gs.patterns, 1, -1 do
        local pattern = gs.patterns[i]
        if pattern.level_done then
            table.remove(gs.patterns, i)
            gs.level_scene:remove(pattern.main_scene)
        else
            local frames = pattern.level_end_frame - pattern.frame_offset
            local t = frames * settings.time_per_frame - pattern.keyframe_time
            if t > end_t then
                end_t = t
            end
        end
    end
    if end_t <= settings.inf_pattern_lead_time then
        start_next_pattern(gs)
    end
end

local point_nodes
local points_group = am.group()
local next_available_point_node = 1
local num_point_nodes = 20

local
function get_point_node()
    if not point_nodes then
        point_nodes = {}
        points_group:remove_all()
        for i = 1, 20 do
            local node = am.translate(0, 0) ^ am.scale(1) ^ am.text(score_font, "0000000") 
            node.hidden = true
            point_nodes[i] = node
            points_group:append(node)
        end
        next_available_point_node = 1
    end
    local node = point_nodes[next_available_point_node]
    node:cancel("ani")
    node.hidden = true
    next_available_point_node = next_available_point_node % num_point_nodes + 1
    return node
end
local points_scale = 0.25
local points_offset = vec2(0, 20)

local
function reset_point_nodes(gs)
    if not point_nodes then
        return
    end
    for _, node in ipairs(point_nodes) do
        node:cancel("ani")
        node.hidden = true
        node"scale".scale2d = vec2(points_scale)
        node"text".color = vec4(gs.palette[6], 1)
    end
end

local
function show_token_points(gs)
    gs.token_points_node = get_point_node()
    gs.token_points_node"scale".scale2d = vec2(0)
    gs.token_points_node"text".color = vec4(gs.palette[6], 1)
    gs.token_points_node:action("ani", am.series{
        am.tween(gs.token_points_node"scale", 0.2, {scale2d = vec2(points_scale * 1.5)}, am.ease.cubic),
        am.tween(gs.token_points_node"scale", 0.2, {scale2d = vec2(points_scale)}, am.ease.cubic),
    })

    gs.token_points_node"text".text = gs.next_token_points
    gs.token_points_node.hidden = false
    local pos = gs.tokens.tokens[gs.inf_token_row][gs.inf_token_col].pos
    gs.token_points_node.position2d = (gs.points_transform * vec4(pos, 0, 1)).xy + points_offset
end

local level_points = {1, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100}

local
function init_inf_game_state()
    local is_flipped = not is_phone
    local points_transform = is_flipped and math.rotate4(math.pi*0.5) or mat4(1)
    local gs = {
        main_scene = is_flipped and am.rotate(math.pi*0.5) or am.group(),
        level_scene = am.group(),
        patterns = {},
        real_time = 0,
        is_gameover = false,
        mode = "infinite",
        stage_stats = init_stage_stats(),
        time_scale = settings.time_scale,
        rand = am.rand(os.time()),
        stage_num = 1,
        level_num = 1,
        player_transform = mat4(1),
        score = 0,
        last_token_collect_time = 0,
        inf_phase = 0,
        last_phase_change_t = 0,
        next_pattern_palette = 1,
        next_pattern_stage = 1,
        inf_collect_note = 0,
        next_token_points = level_points[1],
        next_token_points_level = 1,
        token_points_multiplier = 1,
        color_transition = 0,
        color_cycle_t = 0,
        is_flipped = is_flipped,
        points_transform = points_transform,
    }
    if global_state.game_speed then
        gs.time_scale = gs.time_scale * (global_state.game_speed * 0.01)
    end
    gs.level = {
        grid_rows = 3,
        grid_cols = 3,
    }
    gs.palette = get_level_palette(1)
    compute_play_area(gs)
    init_background(gs)
    init_grid(gs)
    init_tokens(gs)
    init_explosion(gs)
    gs.main_scene:append(gs.level_scene)
    init_player(gs)
    update_patterns(gs)
    update_grid(gs)
    update_tokens(gs)
    update_player_graphics(gs)

    reset_point_nodes(gs)
    show_token_points(gs)

    gs.score_node = am.translate(win.safe_right - 10, win.safe_top - 10)
        ^ am.scale(score_scale) ^ am.text(score_font, "0", "right", "top")

    return gs
end

local
function set_ortho_P(gs)
    local w, h = gs.play_area_width, gs.play_area_height
    gs.proj_mat = math.ortho(-w/2, w/2, -h/2, h/2, -300, 500)
    gs.top_bind.P = gs.proj_mat
end

local
function handle_resize(gs)
    if win:resized() then
        compute_play_area(gs)
        set_ortho_P(gs)
        if is_mobile then
            gs.close_button.x = win.safe_left + 25
            gs.close_button.y = win.safe_top - 20
        end
    end
end

local
function handle_escape(gs)
    show_title()
end

local
function mixcolors(r, r2, i, pal1, pal2, pal3, pal4)
    local col1 = pal1[i]
    local col2 = pal2[i]
    local col3 = pal3[i]
    local col4 = pal4[i]
    return math.mix(math.mix(col1, col3, r2), math.mix(col2, col4, r2), r)
end

local
function update_colors(gs)
    gs.color_cycle_t = gs.color_cycle_t + am.delta_time
    local p1 = gs.token_points_multiplier
    local p2 = p1
    local r = 0
    if p1 > 4 then
        local ct = math.fract(gs.color_cycle_t * 0.1) * 4
        p1 = math.clamp(math.ceil(ct), 1, 4)
        p2 = p1 % 4 + 1
        r = math.fract(ct)
    end
    local r2 = gs.color_transition
    if global_state.less_flashing then
        p1 = 1
        p2 = 1
        r = 0
        r2 = 0
    end

    local palette1 = palettes[p1 + 8]
    local palette2 = palettes[p2 + 8]
    local palette3 = palettes[p1 + 12]
    local palette4 = palettes[p2 + 12]
    if gs.palette ~= palette1 then
        gs.palette = palette1
        gs.top_bind.explosion_color = palette1[4]
    end

    local star_color = mixcolors(r, r2, 5, palette1, palette2, palette3, palette4)
    local bg_color = mixcolors(r, r2, 4, palette1, palette2, palette3, palette4)
    set_bg_color(star_color, bg_color)

    local grid_color = mixcolors(r, r2, 3, palette1, palette2, palette3, palette4)
    local grid_color4 = vec4(grid_color, 1)
    for _, line in ipairs(gs.grid_overlay:all"line") do
        line.color = grid_color4
    end
    gs.tokens.node.color = grid_color
    gs.tokens.node.shadow_color = bg_color
    gs.token_collect_particles.start_color = grid_color4
    gs.token_collect_particles.end_color = grid_color4

    for _, pattern in ipairs(gs.patterns) do
        for _, layer in ipairs(pattern.layer_states) do
            if layer.node then
                local colors = layer_shaders[layer.shader].colors
                local main_color = mixcolors(r, r2, colors[1], palette1, palette2, palette3, palette4)
                local detail_color1 = mixcolors(r, r2, colors[2], palette1, palette2, palette3, palette4)
                local detail_color2 = mixcolors(r, r2, colors[3], palette1, palette2, palette3, palette4)
                local bind = layer.node"bind"
                bind.main_color = main_color
                bind.detail_color1 = detail_color1
                bind.detail_color2 = detail_color2
            end
        end
    end
end

local
function start_game()
    local gs = init_inf_game_state()
    
    local
    function gameover_action(gs)
        if gs.score > (global_state.best_inf_score or 0) then
            global_state.best_inf_score = gs.score
            gs.new_inf_best = true
        end
        if gs.tutorial_node then
            gs.tutorial_node.hidden = true
        end
        submit_score(gs)
        save_state()
        gs.token_collect_particles.hidden = true
        gs.teleport_particles.hidden = true
        reset_point_nodes(gs)
        gs.score_node"text".text = gs.score
        gs.stage_stats.retries = gs.stage_stats.retries + 1
        play_audio_gameover()
        set_star_speed(0)
        make_explosion(gs)
        return am.parallel{
            am.series{
                am.delay(0.75),
                function()
                    show_inf_gameover(gs)
                    return true
                end
            },
            function()
                if global_state.less_flashing then
                    for _, bind in ipairs(gs.player_color_binds) do
                        bind.color = gs.palette[6]
                    end
                    return true
                else
                    if math.fract(am.frame_time * 7) < 0.5 then
                        for _, bind in ipairs(gs.player_color_binds) do
                            bind.color = gs.palette[6]
                        end
                    else
                        for _, bind in ipairs(gs.player_color_binds) do
                            bind.color = gs.palette[4]
                        end
                    end
                end
            end
        }
    end
    local
    function main_action()
        gs.real_time = gs.real_time + am.delta_time
        if handle_player_jump(gs) then
            gs.stage_stats.jumps = gs.stage_stats.jumps + 1
        end
        update_patterns(gs)
        for _, pattern in ipairs(gs.patterns) do
            local next_keyframe_time = pattern.keyframe_time + am.delta_time * pattern.time_scale
            pattern.player_grid_row = gs.player_grid_row
            pattern.player_grid_col = gs.player_grid_col
            local collide_t = check_player_collision_window(pattern, pattern.keyframe_time, next_keyframe_time)
            if collide_t then
                pattern.keyframe_time = collide_t
                pattern.is_gameover = true
                gs.is_gameover = true
            else
                pattern.keyframe_time = next_keyframe_time
            end
            update_level(pattern)
        end
        update_player_graphics(gs)
        if not gs.is_gameover then
            update_tokens(gs)
            gs.color_transition = (gs.next_token_points_level - 1) / (#level_points - 1)
            local t_since_last_collect = gs.real_time - gs.last_token_collect_time
            local m = gs.token_points_multiplier
            if t_since_last_collect > settings.inf_points_increase_window then
                local decay_t = 1 - (t_since_last_collect - settings.inf_points_increase_window) / settings.inf_points_decay_time
                if decay_t > 0 then
                    if not gs.points_level_hwm then
                        gs.points_level_hwm = gs.next_token_points_level
                    end
                    gs.next_token_points = math.floor(math.mix(level_points[1] * m, level_points[gs.points_level_hwm] * m, decay_t))
                    gs.color_transition = (gs.points_level_hwm - 1) / (#level_points - 1) * decay_t
                    --log("c = %f, h = %d, l = %d, d = %f", gs.color_transition, gs.points_level_hwm, #level_points, decay_t)
                else
                    gs.next_token_points = level_points[1] * m
                    gs.points_level_hwm = nil
                end
                gs.next_token_points_level = 1
            end
            gs.token_points_node"text".text = gs.next_token_points
            if gs.token_collected then
                gs.points_level_hwm = nil
                local last = gs.next_token_points_level == #level_points
                if last then
                    play_audio_portal_open()
                    gs.color_cycle_t = gs.color_cycle_t + 5
                    gs.token_points_multiplier = gs.token_points_multiplier + 1
                    m = gs.token_points_multiplier
                end
                gs.score = gs.score + gs.next_token_points

                --gs.token_points_node"text".color = vec4(gs.palette[3], 1)
                gs.token_points_node"scale".scale2d = vec2(points_scale)
                local points_target_pos
                if win.safe_top == win.top then
                    points_target_pos = vec2(gs.play_area_width*0.5-20, gs.play_area_height*0.5-10)
                else
                    points_target_pos = vec2(gs.play_area_width*0.5-20, gs.play_area_height*0.5-40)
                end
                gs.token_points_node:action("ani", am.series{
                    am.parallel{
                        am.tween(gs.token_points_node"scale", 0.3, {scale2d = vec2(points_scale * 1.5)}, am.ease.out(am.ease.cubic)),
                        am.tween(0.2, {position2d = gs.token_points_node.position2d + vec2(0, 20)}, am.ease.quadratic),
                    },
                    am.parallel{
                        am.tween(0.3, {position2d = points_target_pos}, am.ease.quadratic),
                        am.tween(gs.token_points_node"scale", 0.3, {scale2d = vec2(points_scale)}, am.ease.cubic),
                    },
                    function(node)
                        node.hidden = true
                        gs.score_node"text".text = gs.score
                        gs.score_node:action("pulse", am.series{
                            am.tween(gs.score_node"scale", 0.02, {scale2d = score_scale * 1.4}),
                            am.tween(gs.score_node"scale", 0.1, {scale2d = score_scale}),
                        })
                        return true
                    end
                })

                if last then
                    gs.next_token_points_level = 1
                    gs.next_token_points = level_points[1] * m
                elseif gs.next_token_points_level == 1 and level_points[1] * m == gs.next_token_points then
                    gs.next_token_points_level = 2
                    gs.next_token_points = level_points[gs.next_token_points_level] * m
                elseif gs.real_time - gs.last_token_collect_time < settings.inf_points_increase_window then
                    gs.next_token_points_level = gs.next_token_points_level + 1
                    gs.next_token_points = level_points[gs.next_token_points_level] * m
                else
                    -- decaying
                    for i = 1, #level_points do
                        if level_points[i] * m >= gs.next_token_points then
                            gs.next_token_points = level_points[i] * m
                            gs.next_token_points_level = i
                            break
                        end
                    end
                end

                play_audio_collect(gs.next_token_points_level)

                show_token_points(gs)

                gs.last_token_collect_time = gs.real_time
                gs.token_collected = false
            end
        end
        for _, pattern in ipairs(gs.patterns) do
            update_level_node_transforms(pattern)
        end
        update_grid(gs)
        if screenshot_mode then
            if win:key_down"t" then
                manual_t = math.fract(manual_t + am.delta_time * 0.1)
            end
            gs.top_bind.t = manual_t
        else
            gs.top_bind.t = gs.real_time
        end
        if gs.is_gameover then
            check_achievements(gs)
            gs.main_scene:action("main", gameover_action(gs))
            return true
        end
        update_colors(gs)
    end
    gs.main_scene:action("main", main_action)
    gs.top_scene = 
        am.group()
        ^ am.bind{
            P = mat4(1),
            color_offset = vec3(0),
            t = 0,
            explosion_rad = 0,
            explosion_pos = vec2(0),
            explosion_color = gs.palette[4],
            light_dir = math.normalize(vec3(1, 0, 1)),
        }
        ^ {
            gs.main_scene,
            points_group,
        }
    gs.top_bind = gs.top_scene"bind"
    set_ortho_P(gs)

    gs.top_scene:append(gs.score_node)

    if is_mobile then
        local color = vec4(gs.palette[6], 1)
        local close_button = am.translate(win.safe_left + 25, win.safe_top - 20)
            ^ am.scale(0.25) ^ am.sprite(sprites.cross, color)
        gs.top_scene:append(close_button)
        gs.close_button = close_button
    end

    gs.top_scene:action(function() 
        handle_resize(gs)
        if esc_pressed() then
            gs.main_scene:cancel"main"
            handle_escape(gs)
            return true
        end
        if is_mobile and input_pressed() then
            local pos = input_pos()
            if pos.x < win.safe_left + 50 and pos.y > win.safe_top - 50 then
                show_title()
                gs.main_scene:cancel"main"
                play_audio_gameover()
                return true
            end
        end
    end)
    root_scene:remove_all()
    root_scene:append(gs.top_scene)
    main_action()

    if is_mobile and (global_state.best_inf_score or 0) <= 3 then
        show_tutorial(gs)
    end

    start_music(get_stage_first_level(global_state.infinite_music or 1))
end

local
function start_transition(f)
    start_music(get_stage_first_level(global_state.infinite_music or 1))
    local bg = get_bg() 
    local palette = get_level_palette(1)
    local vortex = init_vortex(palette)
    vortex_zoom({palette = palette}, false)
    set_star_speed(800)
    set_bg_color(palette[5], palette[4])
    show_background()
    play_audio_warp()
    root_scene:remove_all()
    root_scene:append(am.group{bg, vortex})
    root_scene:action(am.series{
        am.delay(1),
        function()
            f()
            return true
        end,
    })
end

function start_infinite_game(from_main_menu)
    if from_main_menu then
        -- user could have changed language, so font may have changed
        point_nodes = nil
        start_transition(function()
            start_game()
        end)
    else
        start_game()
    end
end
