require "settings"

local tile_sz = 32
local num_arc_segs = 16
local max_cached_meshes = 50
local cached_level_meshes = {}

local
function search_mesh_cache(name, l)
    local n = #cached_level_meshes
    for i = n, 1, -1 do
        local item = cached_level_meshes[i]
        if item.name == name and item.l == l then
            table.remove(cached_level_meshes, i)
            cached_level_meshes[n] = item
            return item.mesh
        end
    end
    return nil
end

local
function cache_mesh(name, l, mesh)
    local n = #cached_level_meshes
    cached_level_meshes[n + 1] = {name = name, l = l, mesh = mesh}
    if n == max_cached_meshes then
        table.remove(cached_level_meshes, 1)
    end
end

local stencil_vshader = [[
    precision highp float;
    attribute vec2 vert;
    uniform mat4 MV;
    uniform mat4 P;
    void main() {
        gl_Position = P * MV * vec4(vert, 0.0, 1.0);
    }
]]

local stencil_fshader = [[
    precision mediump float;
    void main() {
        gl_FragColor = vec4(1.0);
    }
]]

local stencil_shader = am.program(stencil_vshader, stencil_fshader)

function compute_layer_tris(layer)
    local tiles = layer.tiles
    local num_rows = #tiles
    local num_columns = num_rows > 0 and #tiles[1] or 0
    local sin, cos, pi = math.sin, math.cos, math.pi

    local tris = {}
    local t = 0
    local y = -num_rows / 2 * tile_sz
    for i = 1, num_rows do
        local x = -num_columns / 2 * tile_sz
        for j = 1, num_columns do
            local tile = tiles[i][j]
            local x1, y1, x2, y2 = x, y, x + tile_sz, y + tile_sz
            if tile == 1 then
                -- full quad
                if j > 1 and tiles[i][j-1] == 1 then
                    -- previous tile in row was also a full quad, so just extend it instead
                    -- of creating new triangles
                    t = t - 12
                    tris[t + 5]  = x2
                    tris[t + 9]  = x2
                    tris[t + 11] = x2
                else
                    tris[t + 1]  = x1
                    tris[t + 2]  = y2
                    tris[t + 3]  = x1
                    tris[t + 4]  = y1
                    tris[t + 5]  = x2
                    tris[t + 6]  = y2
                    tris[t + 7]  = x1
                    tris[t + 8]  = y1
                    tris[t + 9]  = x2
                    tris[t + 10] = y1
                    tris[t + 11] = x2
                    tris[t + 12] = y2
                end
                t = t + 12
            elseif tile == 2 then
                tris[t + 1]  = x1
                tris[t + 2]  = y2
                tris[t + 3]  = x1
                tris[t + 4]  = y1
                tris[t + 5]  = x2
                tris[t + 6]  = y2
                t = t + 6
            elseif tile == 3 then
                tris[t + 1]  = x1
                tris[t + 2]  = y2
                tris[t + 3]  = x1
                tris[t + 4]  = y1
                tris[t + 5]  = x2
                tris[t + 6]  = y1
                t = t + 6
            elseif tile == 4 then
                tris[t + 1]  = x1
                tris[t + 2]  = y1
                tris[t + 3]  = x2
                tris[t + 4]  = y1
                tris[t + 5]  = x2
                tris[t + 6]  = y2
                t = t + 6
            elseif tile == 5 then
                tris[t + 1]  = x2
                tris[t + 2]  = y1
                tris[t + 3]  = x2
                tris[t + 4]  = y2
                tris[t + 5]  = x1
                tris[t + 6]  = y2
                t = t + 6
            elseif tile == 6 then
                local right = pi * 0.5
                local prev_x = x1
                local prev_y = y1
                for s = 1, num_arc_segs do
                    local angle = right * 3 + s / num_arc_segs * right
                    local x = x1 + cos(angle) * tile_sz
                    local y = y2 + sin(angle) * tile_sz
                    tris[t + 1] = x1
                    tris[t + 2] = y2
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 7 then
                local right = pi * 0.5
                local prev_x = x2
                local prev_y = y1
                for s = 1, num_arc_segs do
                    local angle = s / num_arc_segs * right
                    local x = x1 + cos(angle) * tile_sz
                    local y = y1 + sin(angle) * tile_sz
                    tris[t + 1] = x1
                    tris[t + 2] = y1
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 8 then
                local right = pi * 0.5
                local prev_x = x2
                local prev_y = y2
                for s = 1, num_arc_segs do
                    local angle = right + s / num_arc_segs * right
                    local x = x2 + cos(angle) * tile_sz
                    local y = y1 + sin(angle) * tile_sz
                    tris[t + 1] = x2
                    tris[t + 2] = y1
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 9 then
                local right = pi * 0.5
                local prev_x = x1
                local prev_y = y2
                for s = 1, num_arc_segs do
                    local angle = right * 2 + s / num_arc_segs * right
                    local x = x2 + cos(angle) * tile_sz
                    local y = y2 + sin(angle) * tile_sz
                    tris[t + 1] = x2
                    tris[t + 2] = y2
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 10 then
                local right = pi * 0.5
                local prev_x = x2
                local prev_y = y2
                for s = 1, num_arc_segs do
                    local angle = right + s / num_arc_segs * right
                    local x = x2 + cos(angle) * tile_sz
                    local y = y1 + sin(angle) * tile_sz
                    tris[t + 1] = x1
                    tris[t + 2] = y2
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 11 then
                local right = pi * 0.5
                local prev_x = x1
                local prev_y = y2
                for s = 1, num_arc_segs do
                    local angle = right * 2 + s / num_arc_segs * right
                    local x = x2 + cos(angle) * tile_sz
                    local y = y2 + sin(angle) * tile_sz
                    tris[t + 1] = x1
                    tris[t + 2] = y1
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 12 then
                local right = pi * 0.5
                local prev_x = x1
                local prev_y = y1
                for s = 1, num_arc_segs do
                    local angle = right * 3 + s / num_arc_segs * right
                    local x = x1 + cos(angle) * tile_sz
                    local y = y2 + sin(angle) * tile_sz
                    tris[t + 1] = x2
                    tris[t + 2] = y1
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            elseif tile == 13 then
                local right = pi * 0.5
                local prev_x = x2
                local prev_y = y1
                for s = 1, num_arc_segs do
                    local angle = s / num_arc_segs * right
                    local x = x1 + cos(angle) * tile_sz
                    local y = y1 + sin(angle) * tile_sz
                    tris[t + 1] = x2
                    tris[t + 2] = y2
                    tris[t + 3] = prev_x
                    tris[t + 4] = prev_y
                    tris[t + 5] = x
                    tris[t + 6] = y
                    t = t + 6
                    prev_x = x
                    prev_y = y
                end
            end
            x = x + tile_sz
        end
        y = y + tile_sz
    end
    return tris
end

function build_layer_mesh(gs, l, layer)
    if l == 1 then
        return nil
    end
    local name = gs.level.name
    local cached_mesh = search_mesh_cache(name, l)
    if cached_mesh then
        return cached_mesh
    end
    local tris = compute_layer_tris(layer)
    local mesh
    if #tris > 0 then
        mesh = am.bind{
            vert = am.vec2_array(compute_layer_tris(layer))
        } ^ am.draw"triangles"
    else
        mesh = am.group()
    end
    cache_mesh(name, l, mesh)
    return mesh
end

local
function set_layer_transforms_dirty(gs)
    for l, layer in ipairs(gs.layer_states) do
        layer.transform_dirty = true
    end
end

local 
function compute_layer_transform(layer_state)
    if not layer_state.transform_dirty then
        return
    end
    local mat = math.translate4(layer_state.pos + layer_state.origin) *
            math.rotate4(layer_state.angle) *
            math.scale4(layer_state.scale) *
            math.translate4(-layer_state.origin)
    if layer_state.parent then
        compute_layer_transform(layer_state.parent)
        layer_state.transform = layer_state.parent.transform * mat
    else
        layer_state.transform = mat
    end
    layer_state.inverse_transform = math.inverse(layer_state.transform)
    layer_state.transform_dirty = false

    if not layer_state.initial_transform then
        layer_state.initial_transform = layer_state.transform
        layer_state.initial_inverse_transform = layer_state.inverse_transform
    end
end

local
function compute_layer_transforms(layers)
    for _, layer in ipairs(layers) do
        compute_layer_transform(layer)
    end
end

local
function init_layer_states(gs) 
    local layers = gs.level.layers
    gs.layer_states = {}
    for l, layer in ipairs(layers) do
        local num_rows = l == 1 and 0 or #layer.tiles
        local num_cols = l == 1 and 0 or num_rows > 0 and #layer.tiles[1] or 0
        local state = {
            num_rows = num_rows,
            num_columns = num_cols,
            pos = vec2(0),
            origin = vec2(0),
            angle = 0,
            scale = vec2(1),
            next_kf = 1,
            time_offset = 0,
            tiles = layer.tiles,
            keyframes = layer.keyframes,
            stencil = layer.stencil,
            shader = layer.shader,
            transform_dirty = true,
            uv_transform = mat4(1),
        }
        gs.layer_states[l] = state
    end
    for l, layer in ipairs(layers) do
        if layer.parent then
            gs.layer_states[l].parent = gs.layer_states[layer.parent]
        end
        local keyframes = layer.keyframes
        local n = #keyframes
        if n > 0 then
            local offset = keyframes[1].frame
            if not gs.frame_offset or gs.frame_offset > offset then
                gs.frame_offset = offset
            end
            local end_frame = keyframes[n].frame
            if not gs.level_end_frame or gs.level_end_frame < end_frame then
                gs.level_end_frame = end_frame
            end
        end
    end
    if not gs.frame_offset then
        gs.frame_offset = 1
    end
    if not gs.level_end_frame then
        gs.level_end_frame = 1
    end
end

local
function stencil_required(gs)
    for _, layer in ipairs(gs.level.layers) do
        if layer.stencil then
            return true
        end
    end
    return false
end

function init_level(gs)
    gs.keyframe_time = 0
    if not gs.layers_group then
        gs.layers_group = am.group()
        gs.main_scene:append(gs.layers_group)
    end
    if stencil_required(gs) then
        gs.stencil_group = am.group()
        gs.layers_group:append(
            am.stencil_test{
                enabled = true,
                ref = 0,
                read_mask = 1,
                write_mask = 1,
                func_front = "never",
                op_fail_front = "invert",
                op_zfail_front = "invert",
                op_zpass_front = "invert",
                func_back = "never",
                op_fail_back = "invert",
                op_zfail_back = "invert",
                op_zpass_back = "invert",
            }
            ^ am.use_program(stencil_shader)
            ^ gs.stencil_group
        )
    end
    init_layer_states(gs)
    init_level_visuals(gs)
end

local
function binary_ease(t)
    if t < 1 then
        return 0
    else 
        return 1
    end
end

layer_easings = {
    am.ease.linear,
    am.ease.quadratic,
    am.ease.cubic,
    am.ease.out(am.ease.quadratic),
    am.ease.out(am.ease.cubic),
    am.ease.sine,
    am.ease.inout(am.ease.quadratic, am.ease.quadratic),
    am.ease.inout(am.ease.cubic, am.ease.cubic),
    binary_ease,
}

local
function setup_next_timed_sublevel(gs)
    gs.compound_level_index = gs.compound_level_index + 1
    gs.level = gs.compound_level.sublevels[gs.compound_level_index].level
    gs.layers_group:remove_all()
    gs.layer_states = nil
    gs.frame_offset = nil
    gs.stencil_group = nil
    init_level(gs)
    update_level(gs)
    update_level_node_transforms(gs)
end

function update_level(gs)
    if gs.level_done then
        return
    end
    local frame_offset = gs.frame_offset
    local t = gs.keyframe_time
    local layers = gs.layer_states
    if gs.new_layer_animation_cycle then
        update_layer_uv_transforms(gs)
        gs.new_layer_animation_cycle = false
    end
    local all_finished = true
    set_layer_transforms_dirty(gs)
    for l, layer in ipairs(layers) do
        local ft = settings.time_per_frame
        local keyframes = layer.keyframes
        local next_kf = layer.next_kf - 1
        local next_t
        local next_keyframe
        repeat
            next_kf = next_kf + 1
            next_keyframe = keyframes[next_kf]
            next_t = next_keyframe and (next_keyframe.frame - frame_offset) * ft
        until not next_keyframe or next_t >= t - layer.time_offset
        layer.next_kf = next_kf
        local prev_keyframe = keyframes[next_kf - 1]
        if next_keyframe and prev_keyframe then
            local ease = layer_easings[prev_keyframe.ease]
            local prev_t = (prev_keyframe.frame - frame_offset) * ft
            local r = ease((t - layer.time_offset - prev_t) / (next_t - prev_t))
            local one_minus_r = 1 - r
            local x = one_minus_r * prev_keyframe.x + r * next_keyframe.x
            local y = one_minus_r * prev_keyframe.y + r * next_keyframe.y
            local angle = one_minus_r * prev_keyframe.angle + r * next_keyframe.angle
            local scale_x = one_minus_r * prev_keyframe.scale_x + r * next_keyframe.scale_x
            local scale_y = one_minus_r * prev_keyframe.scale_y + r * next_keyframe.scale_y
            local origin_x = one_minus_r * prev_keyframe.origin_x + r * next_keyframe.origin_x
            local origin_y = one_minus_r * prev_keyframe.origin_y + r * next_keyframe.origin_y
            local pos = vec2(x, y)
            local scale = vec2(scale_x, scale_y)
            local origin = vec2(origin_x, origin_y)
            layer.pos = pos
            layer.angle = angle
            layer.scale = scale
            layer.origin = origin
        elseif next_keyframe then
            layer.pos = vec2(next_keyframe.x, next_keyframe.y)
            layer.angle = next_keyframe.angle
            layer.scale = vec2(next_keyframe.scale_x, next_keyframe.scale_y)
            layer.origin = vec2(next_keyframe.origin_x, next_keyframe.origin_y)
        elseif prev_keyframe then
            layer.pos = vec2(prev_keyframe.x, prev_keyframe.y)
            layer.angle = prev_keyframe.angle
            layer.scale = vec2(prev_keyframe.scale_x, prev_keyframe.scale_y)
            layer.origin = vec2(prev_keyframe.origin_x, prev_keyframe.origin_y)
        end
        if next_keyframe then
            all_finished = false
        end
    end
    compute_layer_transforms(layers)
    gs.level_cycle_complete = all_finished
    if all_finished then
        if gs.compound_level then
            -- next sublevel
            if gs.compound_level_index == #gs.compound_level.sublevels then
                gs.is_stage_cleared = true
            else
                setup_next_timed_sublevel(gs)
            end
        else
            if gs.mode == "infinite" then
                gs.level_done = true
            else
                -- loop
                for l, layer in ipairs(layers) do
                    layer.next_kf = 1
                    --layer.time_offset = t
                end
                gs.keyframe_time = 0
                gs.new_layer_animation_cycle = true
            end
        end
    end
    gs.player_transform = gs.layer_states[1].transform
    update_level_visuals(gs)
end

function update_level_node_transforms(gs)
    local layers = gs.layer_states
    for l = 2, #layers do
        local layer = layers[l]
        layer.transform_node.mat = layer.transform
    end
end

local
function pos_to_layer_space(pos, layer)
    return (layer.inverse_transform * vec4(pos, 0, 1)).xy
end

local
function layer_collides(pos, layer)
    local num_rows = layer.num_rows
    local num_columns = layer.num_columns
    local layer_pos = pos_to_layer_space(pos, layer)
    local column_frac = layer_pos.x / tile_sz + num_columns / 2
    local column = math.ceil(column_frac)
    local x = 1 - (column - column_frac)
    local row_frac = layer_pos.y / tile_sz + num_rows / 2
    local row = math.ceil(row_frac)
    local y = 1 - (row - row_frac)
    if column < 1 or row < 1 or column > num_columns or row > num_rows then
        return false
    end
    local tile = layer.tiles[row][column]
    --log("%f %f", x, y)
    if tile == 0 then
        return false
    elseif tile == 1 then
        return true
    elseif tile == 2 then
        return x - y < 0
    elseif tile == 3 then
        return 1 - x - y > 0
    elseif tile == 4 then
        return x - y > 0
    elseif tile == 5 then
        return 1 - x - y < 0
    elseif tile == 6 then
        return y > 1 - (1 - x ^ 2) ^ 0.5
    elseif tile == 7 then
        return y < (1 - x ^ 2) ^ 0.5
    elseif tile == 8 then
        return y < (1 - (x - 1) ^ 2) ^ 0.5
    elseif tile == 9 then
        return y > 1 - (1 - (x - 1) ^ 2) ^ 0.5
    elseif tile == 10 then
        return y > (1 - (x - 1) ^ 2) ^ 0.5
    elseif tile == 11 then
        return y < 1 - (1 - (x - 1) ^ 2) ^ 0.5
    elseif tile == 12 then
        return y < 1 - (1 - x ^ 2) ^ 0.5
    elseif tile == 13 then
        return y > (1 - x ^ 2) ^ 0.5
    end
    return false
end

local
function compute_layer_collisions(pos, radius, layer)
    layer.collides1 = layer_collides(pos + vec2(radius, 0), layer)
    layer.collides2 = layer_collides(pos + vec2(-radius, 0), layer)
    layer.collides3 = layer_collides(pos + vec2(0, radius), layer)
    layer.collides4 = layer_collides(pos + vec2(0, -radius), layer)
end

function check_player_collision_window(gs, t1, t2)
    local track = gs.level.collision_windows[gs.player_grid_row][gs.player_grid_col]
    local pre_trim = settings.collision_pre_window * gs.time_scale
    local post_trim = settings.collision_post_window * gs.time_scale
    for i, window in ipairs(track) do
        --print(string.format("%f - %f", t1, t2))
        --print(table.tostring(track))
        if t2 >= (window.start_t or -100000) and t1 < (window.end_t or 100000) then
            local trimmed_start = -100000
            local trimmed_end = 100000
            if window.start_t and window.end_t then
                local mid_t = (window.end_t - window.start_t) * 0.5 + window.start_t
                trimmed_start = math.min(mid_t, window.start_t + pre_trim)
                trimmed_end = math.max(mid_t, window.end_t - post_trim)
            elseif window.start_t then
                trimmed_start = window.start_t + pre_trim
            elseif window.end_t then
                trimmed_end = window.end_t - post_trim
            end
            if t2 >= trimmed_start and t1 < trimmed_end then
                if t1 > (window.start_t or -100000) then
                    return t1
                elseif t2 < (window.end_t or 100000) then
                    return t2
                else
                    return window.start_t
                end
            end
        end
    end
    return nil
end

local
function pos_collides(gs, pos)
    local rad = settings.hit_radius
    local count1, count2, count3, count4 = 0, 0, 0, 0
    local layers = gs.layer_states
    for l = 2, #layers do
        local layer = layers[l]
        compute_layer_collisions(pos, rad, layer)
        count1 = count1 + (layer.collides1 and 1 or 0)
        count2 = count2 + (layer.collides2 and 1 or 0)
        count3 = count3 + (layer.collides3 and 1 or 0)
        count4 = count4 + (layer.collides4 and 1 or 0)
    end
    if count1 == 0 or count2 == 0 or count3 == 0 or count4 == 0 then
        return false
    end
    if not gs.stencil_group then
        return true
    end
    local val1 = count1 % 2
    local val2 = count2 % 2
    local val3 = count3 % 2
    local val4 = count4 % 2
    local pass1, pass2, pass3, pass4 = false, false, false, false
    for l = 2, #layers do
        local layer = layers[l]
        local stencil = layer.stencil
        if not stencil then
            pass1 = pass1 or layer.collides1
            pass2 = pass2 or layer.collides2
            pass3 = pass3 or layer.collides3
            pass4 = pass4 or layer.collides4
        elseif stencil == "zero" then
            pass1 = pass1 or layer.collides1 and val1 == 0
            pass2 = pass2 or layer.collides2 and val2 == 0
            pass3 = pass3 or layer.collides3 and val3 == 0
            pass4 = pass4 or layer.collides4 and val4 == 0
        elseif stencil == "one" then
            pass1 = pass1 or layer.collides1 and val1 == 1
            pass2 = pass2 or layer.collides2 and val2 == 1
            pass3 = pass3 or layer.collides3 and val3 == 1
            pass4 = pass4 or layer.collides4 and val4 == 1
        end
    end
    --log("count: %d %d %d %d  val: %d %d %d %d   pass: %s %s %s %s", count1, count2, count3, count4, val1, val2, val3, val4, pass1, pass2, pass3, pass4)
    return pass1 and pass2 and pass3 and pass4
end

function clear_mesh_cache()
    cached_level_meshes = {}
end

local
function compute_keyframes_total_time(keyframes)
    local n = #keyframes
    if n == 0 then
        return 0
    end
    local frames = keyframes[n].frame - keyframes[1].frame
    return frames * settings.time_per_frame
end

local
function compute_level_time(level)
    local t = 0
    for _, layer in ipairs(level.layers) do
        t = math.max(t, compute_keyframes_total_time(layer.keyframes))
    end
    return t
end

local
function compute_compound_level_time(compound_level)
    local t = 0
    for _, sublevel in ipairs(compound_level.sublevels) do
        t = t + compute_level_time(sublevel.level)
    end
    return t
end

function init_level_progress_bar(gs)
    local total_t = compute_compound_level_time(gs.compound_level)
    gs.compound_level_total_time = total_t
    local node = am.translate(0, gs.play_area_height/2 - 20)
    local w = settings.progress_bar_width
    node:append(am.rect(-w/2, -1, w/2, 1, vec4(gs.palette[3], 1)))
    node:append(am.rect(-w/2, -6, -w/2+2, 6, vec4(gs.palette[3], 1)))
    node:append(am.rect(w/2-2, -6, w/2, 6, vec4(gs.palette[3], 1)))
    local indicator = am.translate(-w/2, 0) ^ am.rotate(math.rad(45)) ^ am.rect(-5, -5, 5, 5, vec4(gs.palette[6], 1))
    node:append(indicator)
    gs.progress_bar_indicator = indicator
    gs.progress_bar = node
    gs.main_scene:append(node)
end

function compute_level_collision_windows(level)
    local total_t = compute_level_time(level)
    local gs = init_game_state(level, 2)

    local dt = settings.sim_time_step
    local windows = {}
    local grid_rows = level.grid_rows
    local grid_cols = level.grid_cols
    for i = 1, grid_rows do
        windows[i] = {}
        for j = 1, grid_cols do
            windows[i][j] = {}
        end
    end

    local positions = gen_grid_positions(grid_rows, grid_cols)
    while not gs.level_cycle_complete do
        for i = 1, grid_rows do
            for j = 1, grid_cols do
                local player_pos = positions[i][j]
                local transformed_pos = (gs.player_transform * vec4(player_pos, 0, 1)).xy
                local collides = pos_collides(gs, transformed_pos)
                local track = windows[i][j]
                local n = #track
                if collides then
                    if n == 0 then
                        if gs.keyframe_time > 0 then
                            table.insert(track, {start_t = gs.keyframe_time})
                        else
                            table.insert(track, {})
                        end
                    elseif track[n].end_t then
                        table.insert(track, {start_t = gs.keyframe_time})
                    end
                else
                    if n > 0 and not track[n].end_t then
                        track[n].end_t = gs.keyframe_time
                    end
                end
            end
        end
        gs.keyframe_time = gs.keyframe_time + dt
        update_level(gs)
    end

    level.collision_windows = windows
end
