local sprites = require "sprites"
local palettes = require "palettes"
require "level_order"
require "level"

function get_level_palette(level_num)
    return palettes[(get_level_stage(level_num) - 1) % #palettes + 1]
end

function get_menu_palette()
    --log(table.tostring(global_state.saved_game))
    return palettes[global_state.saved_game and get_level_stage(global_state.saved_game.level) or 1]
end

local layer_vshader1 = [[
    precision highp float;
    attribute vec2 vert;
    uniform mat4 MV;
    uniform mat4 P;
    uniform mat4 UV;
    varying vec2 v_uv;
    varying vec2 v_pos;
    void main() {
        mat4 MVUV = mat4(
            vec4(length(MV[0].xyz), 0.0, 0.0, 0.0),
            vec4(0.0, length(MV[1].xyz), 0.0, 0.0),
            vec4(0.0, 0.0, length(MV[2].xyz), 0.0),
            vec4(0.0, 0.0, 0.0, 1.0));
        v_uv = (UV * MVUV * vec4(vert, 0.0, 1.0)).xy * 0.012;
        vec4 mvpos = MV * vec4(vert, 0.0, 1.0);
        v_pos = mvpos.xy;
        gl_Position = P * mvpos;
    }
]]

local layer_vshader2 = [[
    precision highp float;
    attribute vec2 vert;
    uniform mat4 MV;
    uniform mat4 P;
    uniform mat4 UV;
    varying vec2 v_uv;
    varying vec2 v_pos;
    void main() {
        vec4 mvpos = MV * vec4(vert, 0.0, 1.0);
        v_pos = mvpos.xy;
        v_uv = v_pos * 0.012;
        gl_Position = P * mvpos;
    }
]]

local layer_fshader = [[
    precision mediump float;
    uniform vec3 main_color;
    uniform vec3 detail_color1;
    uniform vec3 detail_color2;
    uniform float t;
    uniform float speed;
    uniform vec3 explosion_color;
    uniform vec2 explosion_pos;
    uniform float explosion_rad;
    uniform sampler2D tex;
    varying vec2 v_uv;
    varying vec2 v_pos;
    void main() {
        vec4 sample = texture2D(tex, v_uv);
        float r = fract(sample.r - t * speed);
        float r1 = min(0.5, r) * 2.0;
        float r2 = (max(0.5, r) - 0.5) * 2.0;
        vec3 col = mix(main_color, mix(mix(main_color, detail_color1, r1), detail_color2, r2), sample.a);
        vec3 final = mix(col, explosion_color, step(distance(v_pos, explosion_pos), explosion_rad));
        gl_FragColor = vec4(clamp(final, 0.0, 1.0), 1.0);
    }
]]

local layer_shader1 = am.program(layer_vshader1, layer_fshader)
local layer_shader2 = am.program(layer_vshader2, layer_fshader)

layer_shaders = {
    {
        name = "tech",
        program = layer_shader1,
        params = {
            speed = 2,
        },
        colors = {1, 3, 2},
    },
    {
        name = "plasma",
        program = layer_shader2,
        params = {
            speed = 4,
        },
        colors = {17, 19, 18},
    },
}

local
function load_textures()
    for _, shader in ipairs(layer_shaders) do
        shader.textures = {}
        for i = 1, 4 do
            local file = "panel_textures/"..shader.name..i..".png"
            shader.textures[i] = am.texture2d(file)
            shader.textures[i].wrap = "repeat"
        end
    end
end

load_textures()

local
function create_layer_node(gs, layer)
    local shader = layer_shaders[layer.shader]
    local texture_id = (gs.stage_num - 1) % #shader.textures + 1
    local texture = shader.textures[texture_id]
    local colors = shader.colors
    local params = {
        main_color = gs.palette[colors[1]],
        detail_color1 = gs.palette[colors[2]],
        detail_color2 = gs.palette[colors[3]],
        tex = texture,
        UV = mat4(1),
    }
    table.merge(params, shader.params)
    if global_state.less_flashing then
        params.speed = params.speed * 0.5
    end
    layer.node =
        am.bind(params)
        ^ am.use_program(shader.program)
        ^ am.transform(mat4(1))
        ^ layer.mesh
end

function init_level_visuals(gs)
    for l, layer in ipairs(gs.layer_states) do
        layer.mesh = build_layer_mesh(gs, l, layer)
    end
    for l, layer in ipairs(gs.layer_states) do
        if l > 1 then
            create_layer_node(gs, layer)
            layer.transform_node = layer.node"transform"
            if gs.stencil_group then
                if layer.stencil then
                    gs.stencil_group:append(layer.transform_node)
                    local func = "always"
                    local ref = 0
                    if layer.stencil == "zero" then
                        func = "equal"
                    elseif layer.stencil == "one" then
                        func = "equal"
                        ref = 1
                    elseif layer.stencil == "never" then
                        func = "never"
                    end
                    layer.node = 
                        am.stencil_test{
                            enabled = true,
                            ref = ref,
                            read_mask = 1,
                            write_mask = 0,
                            func_front = func,
                            op_fail_front = "keep",
                            op_zfail_front = "keep",
                            op_zpass_front = "keep",
                            func_back = func,
                            op_fail_back = "keep",
                            op_zfail_back = "keep",
                            op_zpass_back = "keep",
                        }
                        ^ layer.node
                end
            end
            gs.layers_group:append(layer.node)
        end
    end
end

function update_layer_uv_transforms(gs)
    local layers = gs.layer_states
    for l = 2, #layers do
        local layer = layers[l]
        local invdelta = layer.initial_transform * layer.inverse_transform
        layer.uv_transform = invdelta * layer.uv_transform
        layer.node.UV = layer.uv_transform
    end
end

function update_level_visuals(gs)
    --if gs.stage_num == 4 then
    --    local gradient_loop_t = 2
    --    local n_colors = 5
    --    local t = (gs.real_time % gradient_loop_t) / gradient_loop_t
    --    local n = t * n_colors
    --    local palette = gs.palette
    --    t = math.fract(n)
    --    n = math.clamp(math.ceil(n), 1, n_colors)
    --    local colora1 = palette[17 + n % n_colors]
    --    local colora2 = palette[17 + (n+1) % n_colors]
    --    local colorb1 = palette[17 + (n+3) % n_colors]
    --    local colorb2 = palette[17 + (n+4) % n_colors]
    --    local colora = math.mix(colora1, colora2, t)
    --    local colorb = math.mix(colorb1, colorb2, t)
    --    local layers = gs.layer_states
    --    for l = 2, #layers do
    --        local layer = layers[l]
    --        layer.node"bind".detail_color1 = colora
    --        layer.node"bind".detail_color2 = colorb
    --        layer.node"bind".main_color = colora * 0.2 + 0.1
    --    end
    --end
end
