require "settings"
require "level_visuals"
require "vortex"
local sprites = require "sprites"

local token_vshader = [[
    precision highp float;
    attribute vec3 pos;
    attribute vec3 norm;
    uniform mat4 MV;
    uniform mat4 P;
    uniform vec3 color;
    uniform vec3 shadow_color;
    uniform vec3 light_dir;
    varying vec3 v_color;
    void main() {
        float r = max(0.0, dot(light_dir, normalize((MV * vec4(norm, 0.0)).xyz)));
        v_color = mix(shadow_color, color, pow(r, 0.5));
        gl_Position = P * MV * vec4(pos, 1.0);
    }
]]

local token_fshader = [[
    precision mediump float;
    varying vec3 v_color;
    void main() {
        gl_FragColor = vec4(v_color, 1.0);
    }
]]

local token_shader = am.program(token_vshader, token_fshader)

local ship_vshader = [[
    precision highp float;
    attribute vec3 pos;
    uniform mat4 MV;
    uniform mat4 P;
    void main() {
        gl_Position = P * MV * vec4(pos, 1.0);
    }
]]

local ship_fshader = [[
    precision mediump float;
    uniform vec3 color;
    void main() {
        gl_FragColor = vec4(color, 1.0);
    }
]]

local token_collect_particles = am.particles2d{
    source_pos = vec2(0),
    source_pos_var = vec2(0),
    max_particles = 200,
    emission_rate = 0,
    start_particles = 0,
    life = 0.25,
    life_var = 0.10,
    angle = 0,
    angle_var = math.rad(180),
    speed = 400,
    speed_var = 300,
    start_color = vec4(1),
    start_color_var = vec4(0),
    end_color = vec4(1),
    end_color_var = vec4(0),
    start_size = 15,
    start_size_var = 5,
    end_size = 3,
    damping = 8,
    sprite_source = sprites.circle,
}
local token_collect_particles_node = am.blend"premult" ^ token_collect_particles

local ship_shader = am.program(ship_vshader, ship_fshader)

local
function load_token_model(name)
    local buf, stride, normals_offset = am.load_obj(name)
    local verts = buf:view("vec3", 0, stride)
    local normals = buf:view("vec3", normals_offset, stride)
    return 
        am.bind{
            norm = normals,
            pos = verts,
        }
        ^am.depth_test("less", true)
        ^am.use_program(token_shader)
        ^am.draw("triangles")
end

local
function load_ship_model(name)
    local buf, stride, uvs_offset = am.load_obj(name)
    local verts = buf:view("vec3", 0, stride)
    return 
        am.bind{
            pos = verts,
        }
        ^am.depth_test("less", true)
        ^am.use_program(ship_shader)
        ^am.draw("triangles")
end

local token_model = load_token_model("models/token.obj")
local ship_inner1_model = load_ship_model("models/ship_inner1.obj")
local ship_inner2_model = load_ship_model("models/ship_inner2.obj")
local ring1_model = load_ship_model("models/ring1.obj")
local ring2_model = load_ship_model("models/ring2.obj")

function get_token_model()
    return token_model
end

local grid_dx = 96
local grid_dy = grid_dx

function gen_grid_positions(num_rows, num_cols)
    local positions = {}
    local w = (num_cols - 1) * grid_dx
    local h = (num_rows - 1) * grid_dy
    local x = -w/2
    local y = h/2
    for i = 1, num_rows do
        positions[i] = {}
        for j = 1, num_cols do
            positions[i][j] = vec2(x, y)
            x = x + grid_dx
        end
        x = -w/2
        y = y - grid_dy
    end
    return positions
end

local
function init_grid_positions(gs)
    local num_rows = gs.level.grid_rows
    local num_cols = gs.level.grid_cols
    gs.grid_positions = gen_grid_positions(num_rows, num_cols)
    gs.grid_center_row = math.ceil(num_rows / 2)
    gs.grid_center_col = math.ceil(num_cols / 2)
    gs.grid_rows = num_rows
    gs.grid_cols = num_cols
end

local grid_line_thickness = 2
local grid_lines = {
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
    am.line(vec2(0), vec2(0), grid_line_thickness),
}
for _, line in ipairs(grid_lines) do
    line"blend".mode = "off"
end

function init_grid(gs)
    init_grid_positions(gs)
    local overlay = am.group()
    local palette = gs.palette
    local color
    if gs.compound_level then
        color = vec4(palette[10], 1)
    else
        color = vec4(palette[3], 1)
    end
    local rows = gs.grid_rows
    local cols = gs.grid_cols
    for i = 1, rows do
        local y = gs.grid_positions[i][1].y
        local line = grid_lines[i]
        line.point1 = vec2(-10000, y)
        line.point2 = vec2(-400, y)
        line.color = color
        overlay:append(line)
    end
    for i = 1, cols do
        local x = gs.grid_positions[1][i].x
        local line = grid_lines[i+rows]
        line.point1 = vec2(x, 10000)
        line.point2 = vec2(x, 400)
        line.color = color
        overlay:append(line)
    end
    for i = 1, #grid_lines do
        grid_lines[i].hidden = i > rows + cols
    end
        
    gs.grid_overlay = overlay
    gs.main_scene:append(gs.grid_overlay)
end

function update_grid(gs)
    local t = gs.animated_grid and 1 or math.min(1, gs.real_time)
    local l = gs.animated_grid and 1000000 or 600
    local e = 0.01
    local transform = gs.player_transform
    local rows = gs.grid_rows
    local cols = gs.grid_cols
    local w2, h2 = gs.play_area_width * 0.5, gs.play_area_height * 0.5
    for i = 1, rows do
        local y = gs.grid_positions[i][1].y
        local x = math.mix(-400, l, t)
        local line = grid_lines[i]
        if gs.animated_grid then
            local p1 = (transform * vec4(-l, y, 0, 1)).xy
            local p2 = (transform * vec4(x, y, 0, 1)).xy
            if gs.stage_num == 4 and gs.compound_level then
                if math.abs(p1.x) < w2 or math.abs(p2.x) < w2 then
                    p1 = p1{x = -w2}
                    p2 = p2{x = w2}
                end
            end
            line.point1 = p1
            line.point2 = p2
        else
            line.point1 = vec2(-l, y)
            line.point2 = vec2(x, y)
        end
    end
    for i = 1, cols do
        local x = gs.grid_positions[1][i].x
        local y = math.mix(400, -l, t)
        local line = grid_lines[i + rows]
        if gs.animated_grid then
            local p1 = (transform * vec4(x, l, 0, 1)).xy
            local p2 = (transform * vec4(x, y, 0, 1)).xy
            if gs.stage_num == 4 and gs.compound_level then
                if math.abs(p1.y) < h2 or math.abs(p2.y) < h2 then
                    p1 = p1{y = -h2}
                    p2 = p2{y = h2}
                end
            end
            line.point1 = p1
            line.point2 = p2
        else
            line.point1 = vec2(x, l)
            line.point2 = vec2(x, y)
        end
    end
end

local
function create_token(gs, i, j)
    local pos = gs.grid_positions[i][j]
    local is_center = gs.mode ~= "infinite" and i == gs.grid_center_row and j == gs.grid_center_col
    local node
    if is_center then
        node = init_vortex(gs.palette)
    else
        node = am.translate(pos) ^ am.rotate(0, math.normalize(math.randvec3())) ^ am.scale(vec3(6)) ^ token_model
        node.hidden = true
        if gs.mode == "infinite" then
            node:action(function(node)
                if not gs.is_gameover then
                    node"rotate".angle = am.frame_time * 5
                end
            end)
        else
            node:action(am.series{
                am.delay((i + j) * 0.08),
                function(node)
                    node.hidden = false
                    return true
                end,
                function(node)
                    if not gs.is_gameover then
                        node"rotate".angle = am.frame_time * 5
                    end
                    --node"scale".scale2d = vec2(math.sin(am.frame_time * 10) * 0.4 + 1.0)
                end,
            })
        end
    end
    return {
        occupied = true,
        node = node,
        is_center = is_center,
        pos = pos,
    }
end

local
function pick_next_inf_token(gs)
    if gs.inf_token_row then
        gs.tokens.tokens[gs.inf_token_row][gs.inf_token_col].node.hidden = true
    else
        -- set to center initially, so we don't pick the players spawn position
        -- (we'll pick something differen below)
        gs.inf_token_row = gs.grid_center_row
        gs.inf_token_col = gs.grid_center_col
    end
    local row, col
    repeat
        row = gs.rand(1, gs.grid_rows)
        col = gs.rand(1, gs.grid_cols)
    until row ~= gs.inf_token_row or col ~= gs.inf_token_col
    gs.inf_token_row = row
    gs.inf_token_col = col
    gs.tokens.tokens[row][col].node.hidden = false
end

local
function create_tokens(gs)
    local group = am.group()
    local tokens = {}

    for i = 1, gs.grid_rows do
        tokens[i] = {}
        for j = 1, gs.grid_cols do
            local token = create_token(gs, i, j)
            group:append(token.node)
            tokens[i][j] = token
        end
    end

    local node = am.bind{
        color = gs.palette[3],
        shadow_color = gs.palette[4],
    } ^ group

    gs.tokens = {
        tokens = tokens,
        group = group,
        node = node,
    }

    if gs.mode == "infinite" then
        pick_next_inf_token(gs)
    end

end

function init_tokens(gs)
    if gs.compound_level then
        return
    end
    create_tokens(gs)
    gs.tokens_left = gs.grid_rows * gs.grid_cols
    gs.total_tokens = gs.tokens_left
    gs.main_scene:append(gs.tokens.node)
    gs.main_scene:append(token_collect_particles_node)
    local particles_color = vec4(gs.palette[3], 1)
    token_collect_particles.emission_rate = 0
    token_collect_particles.start_color = particles_color
    token_collect_particles.end_color = particles_color
    token_collect_particles:reset()
    gs.token_collect_particles = token_collect_particles
    token_collect_particles.hidden = false
end

function fade_player_out(gs)
    local bg_color = gs.palette[4]
    for _, bind in ipairs(gs.player_color_binds) do
        bind:action(am.tween(0.4, {color = bg_color}))
    end
end

function fade_player_in(gs)
    local bg_color = gs.palette[4]
    for _, bind in ipairs(gs.player_color_binds) do
        local color = bind.color
        bind.color = bg_color
        bind:action(am.tween(0.3, {color = color}))
    end
end

local teleport_particles = am.particles2d{
    source_pos = vec2(0),
    source_pos_var = vec2(0),
    max_particles = 300,
    emission_rate = 0,
    life = 0.3,
    life_var = 0.1,
    angle = 0,
    angle_var = math.pi * 2,
    speed = 120,
    start_color = color,
    start_size = 8,
    start_size_var = 3,
    end_size = 0,
    end_size_var = 0,
    damping = 6,
    sprite_source = sprites.circle,
}
local teleport_particles_node = am.blend"premult" ^ teleport_particles

function init_player(gs)
    local radius = 15
    local inner1_bind = am.bind{
            color = gs.palette[11],
        }
        ^ ship_inner1_model
    local inner2_bind = am.bind{
            color = gs.palette[12],
        } 
        ^ ship_inner2_model
    local player_inner_node = 
        am.rotate(0, vec3(1, 0, 0))
        ^ am.rotate(math.rad(45), vec3(0, 1, 0)) 
        ^ am.scale(vec3(50)) 
        ^ {
            inner1_bind
            ,
            inner2_bind
        }
    local ring1_bind = am.bind{color = gs.palette[14]} ^ ring1_model
    local ring2_bind = am.bind{color = gs.palette[13]} ^ ring2_model
    local ring_model = am.group{ring1_bind, ring2_bind}
    local ring_scale = 94
    local player_node = 
        am.translate(0, 0)
        ^ am.scale(20)
        ^ {
            player_inner_node,
            am.scale(vec3(ring_scale)) ^ am.rotate(0, vec3(1, 0, 0)):tag"rot1"
                ^ ring_model,
            am.scale(vec3(ring_scale)) ^ am.rotate(0, vec3(0, 1, 0)):tag"rot2"
                ^ am.rotate(math.rad(90), vec3(1, 0, 0)) ^ ring_model,
        }
    player_node"scale":action(am.tween(0.35, {scale2d = vec2(0.15)}, am.ease.out(am.ease.hyperbola)))

    local particles_color = vec4(gs.palette[6], 1)
    gs.teleport_particles = teleport_particles
    teleport_particles.hidden = false
    teleport_particles.emission_rate = 0
    teleport_particles.start_color = particles_color
    teleport_particles.end_color = particles_color
    teleport_particles:reset()

    local player_group = am.group{teleport_particles_node, player_node}

    gs.player_grid_row = gs.grid_center_row;
    gs.player_grid_col = gs.grid_center_col;
    gs.player_pos = gs.grid_positions[gs.player_grid_row][gs.player_grid_col]
    gs.player_node = player_node
    gs.player_inner_node = player_inner_node
    gs.player_inner_scale = player_inner_node"scale"
    gs.player_group = player_group
    gs.player_color_binds = {inner1_bind, inner2_bind, ring1_bind, ring2_bind}

    gs.main_scene:append(player_group)

    play_audio_land()

    gs.teleport_anis = {}

    if global_state.less_flashing then
        fade_player_in(gs)
    end

end

local teleport_ring_sprite = am.sprite(sprites.ring)
local teleport_star_sprite = am.sprite(sprites.star2)

local
function show_teleport_particles(gs, pos)
    teleport_particles.emission_rate = 100
    teleport_particles.source_pos = (gs.player_transform * vec4(pos, 0, 1)).xy
    gs.last_teleport_particles_t = gs.real_time
end

local
function add_teleport_ani(gs, p1, p2)
    local color = vec4(gs.palette[6], 1)
    local group = gs.player_group
    local teleport_line = am.line(p1, p2, 8, color)
    group:prepend(teleport_line)
    group:action(am.series{
        am.tween(teleport_line, 0.35, {thickness = 0}),--, am.ease.out(am.ease.quadratic)),
        function()
            group:remove(teleport_line)
            return true
        end
    })
    teleport_star_sprite.color = color
    teleport_ring_sprite.color = color
    local exit_scl1 = am.scale(0.250) ^ teleport_star_sprite
    local exit_scl2 = am.scale(0.250) ^ teleport_star_sprite
    local exit_ring = am.scale(0.1) ^ teleport_ring_sprite
    local exit_ani = am.translate(p1) ^ {exit_scl1, exit_scl2, exit_ring}
    local ani = {
        exit_ani = exit_ani,
        line = teleport_line,
        p1 = p1,
        p2 = p2,
    }
    group:prepend(exit_ani)
    group:action(am.series{
        am.parallel{
            am.tween(exit_scl1, 0.1, {scale2d = vec2(0.15, 0.6)}),
            am.tween(exit_scl2, 0.1, {scale2d = vec2(0.6, 0.15)}),
            am.tween(exit_ring, 0.1, {scale2d = vec2(0.3)}),
        },
        am.parallel{
            am.tween(exit_scl1, 0.2, {scale2d = vec2(0)}),
            am.tween(exit_scl2, 0.2, {scale2d = vec2(0)}),
            am.tween(exit_ring, 0.1, {scale2d = vec2(0)}),
        },
        function()
            group:remove(exit_ani)
            table.remove_all(gs.teleport_anis, ani)
            return true
        end,
    })
    if not screenshot_mode then
        local player_node = gs.player_node
        player_node"scale".scale2d = vec2(
            math.abs(math.sign(p1.x - p2.x)) * 0.1,
            math.abs(math.sign(p1.y - p2.y)) * 0.1)
        player_node"scale":action("jump", am.tween(0.2, {scale2d = vec2(0.15)}))
    end
    table.insert(gs.teleport_anis, ani)
end

local
function update_teleport_anis(gs)
    local transform = gs.player_transform
    for _, ani in ipairs(gs.teleport_anis) do
        local pos1 = (transform * vec4(ani.p1, 0, 1)).xy
        local pos2 = (transform * vec4(ani.p2, 0, 1)).xy
        ani.exit_ani.position2d = pos1
        ani.line.point1 = pos1
        ani.line.point2 = pos2
    end
    if gs.last_teleport_particles_t and (gs.real_time - gs.last_teleport_particles_t > 0.1) then
        teleport_particles.emission_rate = 0
        gs.last_teleport_particles_t = nil
    end
end

local
function do_jump(gs, new_row, new_col)
    local rows = gs.grid_rows
    local cols = gs.grid_cols
    local new_pos = gs.grid_positions[new_row][new_col]
    local old_row, old_col = gs.player_grid_row, gs.player_grid_col
    local int1, int2
    if old_row == 1 and new_row == rows then
        int1 = vec2(new_pos.x, 10000)
        int2 = vec2(new_pos.x, -10000)
    elseif old_row == rows and new_row == 1 then
        int1 = vec2(new_pos.x, -10000)
        int2 = vec2(new_pos.x, 10000)
    elseif old_col == 1 and new_col == cols then
        int1 = vec2(-10000, new_pos.y)
        int2 = vec2(10000, new_pos.y)
    elseif old_col == cols and new_col == 1 then
        int1 = vec2(10000, new_pos.y)
        int2 = vec2(-10000, new_pos.y)
    end
    if int1 then
        add_teleport_ani(gs, gs.player_pos, int1)
        add_teleport_ani(gs, int2, new_pos)
    else
        add_teleport_ani(gs, gs.player_pos, new_pos)
    end
    show_teleport_particles(gs, gs.player_pos)
    gs.player_grid_row = new_row
    gs.player_grid_col = new_col
    gs.player_pos = new_pos
    play_audio_jump()
end

local flipped_dir = {
    up = "right",
    right = "down",
    down = "left",
    left = "up",
}

function handle_player_jump(gs)
    local dir = input_movement()
    if dir then
        if gs.is_flipped then
            dir = flipped_dir[dir]
        end
        local rows = gs.grid_rows
        local cols = gs.grid_cols
        local new_row, new_col = gs.player_grid_row, gs.player_grid_col
        if dir == "left" then
            new_col = (gs.player_grid_col - 2) % cols + 1
        elseif dir == "right" then
            new_col = gs.player_grid_col % cols + 1
        elseif dir == "up" then
            new_row = (gs.player_grid_row - 2) % rows + 1
        elseif dir == "down" then
            new_row = gs.player_grid_row % rows + 1
        end
        do_jump(gs, new_row, new_col)
        if gs.input_recording then
            table.insert(gs.input_recording, {
                time = gs.real_time + gs.frame_offset * settings.time_per_frame,
                x = gs.player_pos.x,
                y = gs.player_pos.y,
            })
        end
        return true
    end
end

local screenshot_mode_rot1 = math.random(10)
local screenshot_mode_rot2 = math.random(10)
local screenshot_mode_inner_angle = math.random(10)

function update_player_graphics(gs)
    if screenshot_mode then
        if win:key_pressed"p" then
            screenshot_mode_rot1 = math.random(10)
            screenshot_mode_rot2 = math.random(10)
            screenshot_mode_inner_angle = math.random(10)
        end
        gs.player_node"rot1".angle = screenshot_mode_rot1
        gs.player_node"rot2".angle = screenshot_mode_rot2
        gs.player_inner_node.angle = screenshot_mode_inner_angle
    else
        gs.player_node"rot1".angle = am.frame_time * 5
        gs.player_node"rot2".angle = am.frame_time * 4.5
        gs.player_inner_node.angle = -am.frame_time * 2
    end
    gs.player_node.position2d = (gs.player_transform * vec4(gs.player_pos, 0, 1)).xy

    update_teleport_anis(gs)
end

local
function show_collect_particles(gs, token)
    token_collect_particles.emission_rate = 400
    token_collect_particles.source_pos = token.node.position2d
    gs.last_token_particles_emit_t = gs.real_time
end

function update_tokens(gs)
    if gs.compound_level then
        return
    end
    if gs.last_token_particles_emit_t and gs.real_time - gs.last_token_particles_emit_t > 0.08 then
        token_collect_particles.emission_rate = 0
        gs.last_token_particles_emit_t = nil
    end
    local token = gs.tokens.tokens[gs.player_grid_row][gs.player_grid_col]
    if gs.mode == "infinite" then
        if gs.player_grid_row == gs.inf_token_row and gs.player_grid_col == gs.inf_token_col then
            pick_next_inf_token(gs)
            global_state.total_cubes_collected = (global_state.total_cubes_collected or 0) + 1
            gs.token_collected = true
            show_collect_particles(gs, token)
        end
    elseif token.occupied and (not token.is_center or gs.tokens_left == 1) then
        token.occupied = false
        if not token.is_center then
            local group = gs.tokens.group
            group:remove(token.node)
            play_audio_collect(gs.total_tokens - gs.tokens_left)
            show_collect_particles(gs, token)
            global_state.total_cubes_collected = (global_state.total_cubes_collected or 0) + 1
        end
        gs.tokens_left = gs.tokens_left - 1
        if gs.tokens_left == 1 then
            vortex_appear(gs)
            play_audio_portal_open()
        end
        if gs.tokens_left == 0 then
            gs.is_stage_cleared = true
        end
    end

    local transform = gs.player_transform
    for i = 1, gs.grid_rows do
        for j = 1, gs.grid_cols do
            local token = gs.tokens.tokens[i][j]
            if token.is_center then
                local p = (transform * vec4(token.pos, 0, 1)).xy
                set_vortex_pos(gs, p)
            else
                token.node.position2d = (transform * vec4(token.pos, 0, 1)).xy
            end
        end
    end
end

function show_ship_hero()
    local palettes = require "palettes"
    local palette = palettes[1]
    local radius = 15
    local inner1_bind = am.bind{
            color = palette[11],
        }
        ^ ship_inner1_model
    local inner2_bind = am.bind{
            color = palette[12],
        } 
        ^ ship_inner2_model
    local player_inner_node = 
        am.rotate(math.rad(10), vec3(1, 0, 0))
        ^ am.rotate(math.rad(60), vec3(0, 1, 0)) 
        ^ am.scale(vec3(50)) 
        ^ {
            inner1_bind
            ,
            inner2_bind
        }
    local ring1_bind = am.bind{color = palette[14]} ^ ring1_model
    local ring2_bind = am.bind{color = palette[13]} ^ ring2_model
    local ring_model = am.group{ring1_bind, ring2_bind}
    local ring_scale = 105
    local player_node = 
        am.depth_test("less", true)
        ^ am.bind{P = math.perspective(math.rad(60), win.width / win.height, 0.1, 1000)}
        ^ am.translate(0, 0, -300)
        ^ {
            player_inner_node,
            am.scale(vec3(ring_scale)) ^ am.rotate(math.rad(15), vec3(1, 0, 0)):tag"rot1"
                ^ ring_model,
            am.scale(vec3(ring_scale)) ^ am.rotate(math.rad(80), vec3(0, 1, 0)):tag"rot2"
                ^ am.rotate(math.rad(90), vec3(1, 0, 0)) ^ ring_model,
        }
    root_scene:remove_all()
    root_scene:append(player_node)
    win.clear_color = vec4(1)
end
