require "window"
local sprites = require "sprites"
local palettes = require "palettes"

local vortex_vshader = [[
    precision highp float;
    attribute vec3 vert;
    uniform mat4 MV;
    uniform mat4 P;
    uniform vec2 curve;
    varying float v_z;
    void main() {
        v_z = vert.z;
        vec4 pos = vec4(vert, 1.0);
        vec4 mvpos = MV * pos;
        if (mvpos.z < -30.0) {
            float t = (-30.0 - mvpos.z) / 50.0;
            mvpos.xy -= pow(t, 3.0) * curve;
        }
        gl_Position = P * mvpos;
    }
]]

local vortex_fshader = [[
    precision mediump float;
    uniform vec3 color1;
    uniform vec3 color2;
    uniform float t;
    uniform float line_width;
    uniform float dz_inv;
    varying float v_z;
    void main() {
        vec3 color = mix(color1, color2, floor(fract(v_z * dz_inv + t) + line_width));
        //vec3 color = mix(color1, color2, floor(fract((v_z * dz_inv + t) * 0.3) + 0.1));
        gl_FragColor = vec4(color, 1.0);
    }
]]

local vortex_shader = am.program(vortex_vshader, vortex_fshader)

local z0 = -15
local dz = 5

local quads
local num_quads

local
function make_quads()
    local depth = 22
    local r = 2
    num_quads = depth * 4
    quads = am.quads(num_quads, {"vert", "vec3", "color", "vec3"})
    local z = -dz * depth
    local x1 = -r
    local x2 = r
    local y1 = -r
    local y2 = r
    for d = 1, depth do
        local z1 = z + dz
        local z2 = z
        -- left
        quads:add_quad{
            vert = {x1, y2, z1,  x1, y1, z1,  x1, y1, z2,  x1, y2, z2},
        }
        -- bottom
        quads:add_quad{
            vert = {x1, y1, z2,  x1, y1, z1,  x2, y1, z1,  x2, y1, z2},
        }
        -- right
        quads:add_quad{
            vert = {x2, y2, z2,  x2, y1, z2,  x2, y1, z1,  x2, y2, z1},
        }
        -- top
        quads:add_quad{
            vert = {x1, y2, z1,  x1, y2, z2,  x2, y2, z2,  x2, y2, z1},
        }
        z = z + dz
    end
end

make_quads()

local node 
local
function make_node()
    node = 
        am.translate(0, 0, z0)
        ^ am.rotate(0)
        ^ am.scale(1)
        ^ am.bind{
            P = math.perspective(math.rad(60), win.width/win.height, 0.1, 1000),
            color1 = vec3(1, 0, 0),
            color2 = vec3(1, 2, 0),
            line_width = 0.3,
            dz_inv = 1/dz * 0.5,
            t = 0,
            curve = vec2(1.5),
        }
        ^ am.use_program(vortex_shader) 
        ^ quads
end

make_node()

local
function appear_ani(gs)
    node.x = 0
    node.y = 0
    node"rotate".angle = 0
    local t = am.frame_time
    node.z = z0
    --local num_elems = num_quads * 6
    local rotate = node"rotate"
    rotate.angle = 0
    local scale = node"scale"
    scale.scale = vec3(0)
    local t = 0
    node:action("ani", function()
        if gs.is_gameover then 
            return true
        end
        t = t + am.delta_time
        node"bind".t = math.fract(am.frame_time * 4)
        node.z = z0 + am.ease.cubic(math.fract(am.frame_time * 2)) * 4 - 4
        rotate.angle = -t
        scale.scale = vec3(math.min(1, t * 3))
    end)
end

local
function set_curve(amt, a)
    a = a or math.random() * math.pi * 2
    node"bind".curve = vec2(math.cos(a), math.sin(a)) * amt
end

function init_vortex(palette)
    node"bind".P = math.perspective(math.rad(60), win.width/win.height, 0.1, 1000)
    node"bind".color1 = palette[4]
    node"bind".color2 = palette[3]
    node"bind".line_width = 0.3
    set_curve(1.5)
    node.hidden = true
    node:cancel"ani"
    return node
end

local vortex_bg_speed = 1

function init_vortex_as_background(gs, darker)
    vortex_bg_speed = 1
    if darker then
        node"bind".color1 = gs.palette[9] * 0.2
        node"bind".color2 = gs.palette[10] * 0.2
    else
        node"bind".color1 = gs.palette[3]
        node"bind".color2 = gs.palette[6]
    end
    node"bind".line_width = 0.3
    node"bind".P = math.perspective(math.rad(60), win.width/win.height, 0.1, 1000)
    set_curve(3.5)
    node"scale".scale = vec3(1)
    node.hidden = false
    node:cancel"ani"
    local rot = node"rotate"
    local t = 0
    node.z = z0 + 20
    local finished_setting_colors = false
    node:action("ani", function()
        if gs.is_gameover then
            return true
        end
        t = t + am.delta_time * vortex_bg_speed
        node.x = math.sin(t * 0.4 + 1)
        node.y = math.sin(t * 0.7)
        rot.angle = -t * 0.3
        node"bind".t = -t * 2
        set_curve(3.5, t * 0.2)
        if not finished_setting_colors then
            if t < 0.5 then
                local r = t / 0.5
                if darker then
                    node"bind".color1 = math.mix(gs.palette[9]*0.2, gs.palette[9], r)
                    node"bind".color2 = math.mix(gs.palette[10]*0.2, gs.palette[10], r)
                else
                    node"bind".color1 = math.mix(gs.palette[3], gs.palette[9], r)
                    node"bind".color2 = math.mix(gs.palette[6], gs.palette[10], r)
                end
            else
                node"bind".color1 = gs.palette[9]
                node"bind".color2 = gs.palette[10]
                finished_setting_colors = true
            end
        end
    end)
    gs.main_scene:append(node)
end

function vortex_zoom_background()
    local t = 0
    node:action("ani", function()
        t = t + am.delta_time
        node.z = z0 + 20 + t * 150
        node"bind".t = t * 6
    end)
end

function vortex_speedup_background()
    vortex_bg_speed = 2
end

function vortex_rainbow()
    local t = 0
    local cycle_t = 1
    local n = 4
    node:action("ani", function()
        t = t + am.delta_time
        node"bind".t = -t * 4
        node"rotate".angle = -t * 0.3
        set_curve(3.5, t * 0.2)
        local ct = math.fract(t / (n * cycle_t)) * n
        local p = math.clamp(math.ceil(ct), 1, n)
        local p0 = p - 1
        if p0 < 1 then
            p0 = n
        end
        local palette = palettes[p]
        local palette0 = palettes[p0]
        local r = math.fract(ct)
        node"bind".color1 = math.mix(palette0[9], palette[9], r)
        node"bind".color2 = math.mix(palette0[10], palette[10], r)
    end)
end

function set_vortex_pos(gs, pos)
    pos = pos / vec2(gs.play_area_width/2, gs.play_area_height/2)
    local translate = math.translate4(vec3(pos, 0))
    local angle = math.rad(60)
    if gs.play_area_height > gs.play_area_width then
        angle = math.atan((gs.play_area_height / settings.min_screen_height) * math.tan(angle * 0.5)) * 2
    end
    node"bind".P = translate * math.perspective(angle, win.width/win.height, 0.1, 1000)
end

function set_vortex_color(color1, color2)
    node"bind".color1 = color1
    node"bind".color2 = color2
end

function vortex_zoom(gs, into_transition)
    node.hidden = false
    local t = 0
    local s1 = math.random(2) * 2 - 3
    local s2 = math.random(2) * 2 - 3
    local r1 = math.random() * 0.5 + 0.25
    local r2 = math.random() * 0.5 + 0.25
    local speed = 120
    local rot = node"rotate"
    node"scale".scale = vec3(1)
    local color_t = 1.0
    local t0
    set_bg_color(gs.palette[5], gs.palette[4])
    node:action("ani", function()
        t = t + am.delta_time
        local move_t = false
        if into_transition then
            local z = z0 + t * speed
            if z > 16 then
                z = 16
                move_t = true
                t0 = t0 or t
            end
            node.z = z
            set_curve(6, t)
        else
            node.z = z0 + t * speed
        end
        node.x = s1 * math.sin(t * 2) * r1
        node.y = s2 * math.sin(t * 2) * r2
        rot.angle = - t * 3
        if move_t then
            if into_transition then
                node"bind".t = -t * math.mix(12, 8, math.clamp((t - t0), 0, 1))
            else
                node"bind".t = -t * 1
            end
        else
            node"bind".t = t * 3.6
        end
        if into_transition then
            if t < color_t then
                local r = t / color_t
                node"bind".color1 = math.mix(gs.palette[4], gs.palette[3], r)
                node"bind".color2 = math.mix(gs.palette[3], gs.palette[6], r)
            else
                node"bind".color1 = gs.palette[9]
                node"bind".color2 = gs.palette[10]
            end
        else
            node"bind".color1 = gs.palette[4]
            if global_state.less_flashing then
                local color1 = math.mix(gs.palette[3], gs.palette[4], 0.5)
                local color2 = gs.palette[4]
                if t < 0.8 then
                    local r = 1 - (1 - (t / 0.8)) ^ 3
                    node"bind".color2 = math.mix(color1, color2, r)
                else
                    node"bind".color2 = color2
                end
            else
                if t < 1.05 then
                    local r = 1 - (1 - (t / 1.05)) ^ 1.5
                    node"bind".color2 = math.mix(gs.palette[3], gs.palette[4], r)
                end
            end
        end
    end)
end

local flash_ring_sprite = am.sprite(sprites.ring)
local flash_star_sprite1 = am.sprite(sprites.star2)
local flash_star_sprite2 = am.sprite(sprites.star2)
local flash_exit_scl1 = am.scale(0.250) ^ flash_star_sprite1
local flash_exit_scl2 = am.scale(0.1) ^ flash_star_sprite2
local flash_exit_ring = am.scale(0.1) ^ flash_ring_sprite
local flash_exit_ani = am.translate(0, 0) ^ {flash_exit_scl1, flash_exit_scl2, flash_exit_ring}

local
function vortex_appear_flash(gs)
    local group = gs.tokens.group
    flash_ring_sprite.color = vec4(gs.palette[6], 1)
    flash_star_sprite1.color = vec4(gs.palette[6], 1)
    flash_star_sprite2.color = vec4(gs.palette[6], 1)
    flash_exit_scl1.scale2d = vec2(0.250)
    flash_exit_scl2.scale2d = vec2(0.1)
    flash_exit_ring.scale2d = vec2(0.1)
    local pos = (gs.player_transform * vec4(gs.grid_positions[gs.grid_center_row][gs.grid_center_col], 0, 1)).xy
    flash_exit_ani.position2d = pos
    group:append(flash_exit_ani)
    group:action("vortex_flash", am.series{
        am.parallel{
            am.tween(flash_exit_scl1, 0.1, {scale2d = vec2(0.2, 1.0)}),
            am.tween(flash_exit_scl2, 0.1, {scale2d = vec2(2.0, 0.1)}),
            am.tween(flash_exit_ring, 0.1, {scale2d = vec2(0.4)}),
        },
        am.parallel{
            am.tween(flash_exit_scl1, 0.2, {scale2d = vec2(0)}),
            am.tween(flash_exit_scl2, 0.2, {scale2d = vec2(0)}),
            am.tween(flash_exit_ring, 0.02, {scale2d = vec2(0)}),
        },
        function()
            group:remove(flash_exit_ani)
            return true
        end
    })
end

function vortex_appear(gs)
    vortex_appear_flash(gs)
    appear_ani(gs)
    node.hidden = false
end
